# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import threading
from abc import ABC, abstractmethod
from typing import Any

from mplang.core.mask import Mask


class ICommunicator(ABC):
    """Base class for communicators."""

    @property
    @abstractmethod
    def rank(self) -> int:
        """Get the rank of this process"""

    @property
    @abstractmethod
    def world_size(self) -> int:
        """Get the world size of this process"""

    @abstractmethod
    def new_id(self) -> str:
        """Must be implemented by mixing class"""
        raise NotImplementedError

    @abstractmethod
    def send(self, to: int, key: str, data: Any) -> None:
        """Send data to peer with the given key"""

    @abstractmethod
    def recv(self, frm: int, key: str) -> Any:
        """Receive data from peer with the given key"""


class ICollective(ABC):
    """Interface for collective communication"""

    @abstractmethod
    def p2p(self, frm: int, to: int, data: Any) -> Any:
        """Perform point-to-point communication"""

    @abstractmethod
    def gather(self, root: int, data: Any) -> list[Any]:
        """Gather data from all processes to root"""

    @abstractmethod
    def gather_m(self, pmask: int, root: int, data: Any) -> list[Any]:
        """Gather data from parties in pmask to root"""

    @abstractmethod
    def scatter(self, root: int, args: list[Any]) -> Any:
        """Scatter data from root to all processes"""

    @abstractmethod
    def scatter_m(self, pmask: int, root: int, args: list[Any]) -> Any:
        """Scatter data from root to parties in pmask"""

    @abstractmethod
    def allgather(self, arg: Any) -> list[Any]:
        """Gather data from all processes to all processes"""

    @abstractmethod
    def allgather_m(self, pmask: int, arg: Any) -> list[Any]:
        """Gather data from parties in pmask to all processes"""

    @abstractmethod
    def bcast(self, root: int, arg: Any) -> Any:
        """Broadcast data from root to all processes"""

    @abstractmethod
    def bcast_m(self, pmask: int, root: int, arg: Any) -> Any:
        """Broadcast data from root to parties in pmask"""


def is_rank_in(rank: int, mask: int) -> bool:
    """Check if the given rank is in the mask"""
    return (1 << rank) & mask != 0


class CollectiveMixin(ICommunicator, ICollective):
    """Mixin class providing default implementations of collective communication algorithms

    This mixin provides implementations based on send/recv primitives.
    Classes using this mixin must implement the ICommunicator interface methods.
    """

    # Note: These will be provided by mixing classes as properties
    @property
    def rank(self) -> int:
        """Must be implemented by mixing class"""
        raise NotImplementedError

    @property
    def world_size(self) -> int:
        """Must be implemented by mixing class"""
        raise NotImplementedError

    def send(self, to: int, key: str, data: Any) -> None:
        """Must be implemented by mixing class"""
        raise NotImplementedError

    def recv(self, frm: int, key: str) -> Any:
        """Must be implemented by mixing class"""
        raise NotImplementedError

    def new_id(self) -> str:
        """Must be implemented by mixing class"""
        raise NotImplementedError

    def p2p(self, frm: int, to: int, data: Any) -> Any:
        """Perform point-to-point communication"""
        # p2p is a special collective operation, with non-sender and non-receiver nodes get None
        assert 0 <= frm < self.world_size
        assert 0 <= to < self.world_size

        cid = self.new_id()

        if self.rank == frm:
            self.send(to, cid, data)

        if self.rank == to:
            return self.recv(frm, cid)
        else:
            return None

    def gather_m(self, pmask: int, root: int, data: Any) -> list[Any]:
        """Gather data from parties in pmask to root"""
        assert 0 <= root < self.world_size
        # wmask = (1 << self.world_size) - 1
        # assert mpt.is_subset(pmask, wmask)

        cid = self.new_id()

        if self.rank in Mask(pmask):
            self.send(root, cid, data)

        if self.rank == root:
            res = [self.recv(idx, cid) for idx in Mask(pmask)]
        else:
            res = [None] * Mask(pmask).num_parties()

        return res

    def gather(self, root: int, data: Any) -> list[Any]:
        """Gather data from all processes to root"""
        pmask = Mask.all(self.world_size)
        return self.gather_m(pmask.value, root, data)

    def scatter_m(self, pmask: int, root: int, args: list[Any]) -> Any:
        """Scatter data from root to parties in pmask"""
        logging.debug(
            f"[{self.rank}]: scatter_m: pmask={pmask}, root={root}, args={args}"
        )
        assert 0 <= root < self.world_size
        mask = Mask(pmask)
        assert len(args) == mask.num_parties(), f"{len(args)} != {mask.num_parties()}"

        cid = self.new_id()

        if self.rank == root:
            for idx, arg in zip(mask, args, strict=True):
                self.send(idx, cid, arg)

        if self.rank in mask:
            data = self.recv(root, cid)
        else:
            data = None

        return data

    def scatter(self, root: int, args: list[Any]) -> Any:
        """Scatter data from root to all processes"""
        pmask = Mask.all(self.world_size)
        return self.scatter_m(pmask.value, root, args)

    def allgather_m(self, pmask: int, arg: Any) -> list[Any]:
        """Gather data from parties in pmask to all parties"""
        logging.debug(f"allgather_m: pmask={pmask}, arg={arg}")
        cid = self.new_id()

        if self.rank in Mask(pmask):
            for idx in Mask(pmask):
                self.send(idx, cid, arg)

            res = [self.recv(idx, cid) for idx in Mask(pmask)]
        else:
            res = [None] * Mask(pmask).num_parties()

        return res

    def allgather(self, arg: Any) -> list[Any]:
        """Gather data from all processes to all processes"""
        pmask = Mask.all(self.world_size)
        return self.allgather_m(pmask.value, arg)

    def bcast_m(self, pmask: int, root: int, arg: Any) -> Any:
        """Broadcast data from root to parties in pmask"""
        logging.debug(f"bcast_m: pmask={pmask}, root={root}, arg={arg}")
        assert 0 <= root < self.world_size

        cid = self.new_id()

        if self.rank == root:
            for idx in Mask(pmask):
                self.send(idx, cid, arg)

        if self.rank in Mask(pmask):
            return self.recv(root, cid)
        else:
            return None

    def bcast(self, root: int, arg: Any) -> Any:
        """Broadcast data from root to all processes"""
        pmask = Mask.all(self.world_size)
        return self.bcast_m(pmask.value, root, arg)


class CommunicatorBase(ICommunicator):
    """Base implementation providing message box functionality for local communication"""

    def __init__(self, rank: int, world_size: int):
        self._rank = rank
        self._world_size = world_size
        self._msgboxes: dict = {}
        self._cond = threading.Condition()
        self._counter = 0

    @property
    def rank(self) -> int:
        return self._rank

    @property
    def world_size(self) -> int:
        return self._world_size

    # override
    def new_id(self) -> str:
        # Ensure thread-safe ID generation
        with self._cond:
            res = self._counter
            self._counter += 1
            return str(res)

    def recv(self, frm: int, key: str) -> Any:
        """Wait until the key is set, returns the value"""
        # print(f"recv {key}: {sender_rank} -> {self.rank}")
        mkey = (frm, key)
        with self._cond:
            # Wait until message arrives, then consume it
            while mkey not in self._msgboxes:
                self._cond.wait()
            return self._msgboxes.pop(mkey)

    def onSent(self, frm: int, key: str, data: Any) -> None:
        """Called when a key is sent to self"""
        with self._cond:
            mkey = (frm, key)
            assert mkey not in self._msgboxes, f"{mkey} exist {self._msgboxes.keys()}"
            self._msgboxes[mkey] = data
            self._cond.notify_all()



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib
import pathlib
import pkgutil
from collections.abc import Callable
from functools import partial, wraps
from types import ModuleType
from typing import Any

from mplang.core.mask import Mask
from mplang.core.mpobject import MPObject
from mplang.core.mptype import Rank
from mplang.core.primitive import (
    constant,
    pconv,
    peval,
    prand,
    prank,
    pshfl,
    pshfl_s,
    uniform_cond,
    while_loop,
)
from mplang.ops import ibis_cc, jax_cc
from mplang.ops.base import FeOperation
from mplang.simp.mpi import allgather_m, bcast_m, gather_m, p2p, scatter_m
from mplang.simp.random import key_split, pperm, prandint, ukey, urandint
from mplang.simp.smpc import reveal, revealTo, seal, sealFrom, srun

# Public exports of the simplified party execution API.
# NOTE: Replaces previous internal __reexport__ (not a Python convention)
# to make star-imports explicit and tooling-friendly.
__all__ = [  # noqa: RUF022
    "MPObject",
    "P",
    "P0",
    "P1",
    "P2",
    "P2P",
    "Party",
    "allgather_m",
    "bcast_m",
    "constant",
    "gather_m",
    "key_split",
    "load_module",
    "p2p",
    "pconv",
    "peval",
    "pperm",
    "prand",
    "prandint",
    "prank",
    "pshfl",
    "pshfl_s",
    "reveal",
    "revealTo",
    "run",
    "runAt",
    "scatter_m",
    "seal",
    "sealFrom",
    "srun",
    "ukey",
    "uniform_cond",
    "urandint",
    "while_loop",
]


def run_impl(
    pmask: Mask | None,
    func: Callable,
    *args: Any,
    **kwargs: Any,
) -> Any:
    """
    Run a function that can be evaluated by the mplang system.

    This function provides a dispatch mechanism based on the first argument
    to route different function types to appropriate handlers.

    Args:
        pmask: The party mask of this function, None indicates auto deduce parties.
        func: The function to be dispatched and executed
        *args: Positional arguments to pass to the function
        **kwargs: Keyword arguments to pass to the function

    Returns:
        The result of evaluating the function through the appropriate handler

    Raises:
        ValueError: If builtin.write is called without required arguments
        TypeError: If the function compilation or evaluation fails
        RuntimeError: If the underlying peval execution encounters errors

    Examples:
        Reading data from a file:

        >>> tensor_info = TensorType(shape=(10, 10), dtype=np.float32)
        >>> attrs = {"format": "binary"}
        >>> result = run_impl(builtin.read, "data/input.bin", tensor_info, attrs)

        Writing data to a file:

        >>> run_impl(builtin.write, data, "data/output.bin")

        Running a JAX function:

        >>> def matrix_multiply(a, b):
        ...     return jnp.dot(a, b)
        >>> result = run_impl(matrix_multiply, mat_a, mat_b)

        Running a custom computation function:

        >>> def compute_statistics(data):
        ...     mean = jnp.mean(data)
        ...     std = jnp.std(data)
        ...     return {"mean": mean, "std": std}
        >>> stats = run_impl(compute_statistics, dataset)
    """

    if isinstance(func, FeOperation):
        pfunc, eval_args, out_tree = func(*args, **kwargs)
    else:
        if ibis_cc.is_ibis_function(func):
            pfunc, eval_args, out_tree = ibis_cc.ibis_compile(func, *args, **kwargs)
        else:
            # unknown python callable, treat it as jax function
            pfunc, eval_args, out_tree = jax_cc.jax_compile(func, *args, **kwargs)
    results = peval(pfunc, eval_args, pmask)
    return out_tree.unflatten(results)


# run :: (a -> a) -> m a -> m a
def run(pyfn: Callable) -> Callable:
    return partial(run_impl, None, pyfn)


# runAt :: Rank -> (a -> a) -> m a -> m a
def runAt(rank: Rank, pyfn: Callable) -> Callable:
    pmask = Mask.from_ranks(rank)
    return partial(run_impl, pmask, pyfn)


def P2P(src: Party, dst: Party, value: Any) -> Any:
    """Point-to-point transfer using Party objects instead of raw ranks.

    Equivalent to ``p2p(src.rank, dst.rank, value)`` but improves readability
    and reduces magic numbers in user code / tutorials.

    Parameters
    ----------
    src : Party
        Source party object.
    dst : Party
        Destination party object.
    value : Any
        Value to transfer.

    Returns
    -------
    Any
        The same value representation at destination context (as defined by
        underlying ``p2p`` primitive semantics).
    """
    if not isinstance(src, Party) or not isinstance(dst, Party):  # defensive
        raise TypeError("P2P expects Party objects, e.g. P2P(P0, P2, value)")
    return p2p(src.rank, dst.rank, value)


"""Party-scoped module registration & dispatch.

This module provides a light-weight mechanism to expose *module-like* groups
of callable operations bound to a specific party (rank) via attribute access:

    load_module("mplang.ops.crypto", alias="crypto")
    P0.crypto.encrypt(x)  # executes encrypt() with pmask = {rank 0}

Core concepts:
* Registry (``_NAMESPACE_REGISTRY``): maps alias -> importable module path.
* Lazy import: underlying module is imported on first attribute access.
* Wrapping: fetched callables are wrapped so that invocation automatically
    routes through ``run_impl`` with that party's mask.

Only *callable* attributes are exposed; non-callable attributes raise
``AttributeError`` to avoid surprising divergent local vs. distributed
semantics.

The public API surface intentionally stays small (`Party`, `P`, `run`,
`runAt`, and `load_module`). Internal details (proxy class / registry) are
not part of the stability guarantee.
"""

_NAMESPACE_REGISTRY: dict[str, str] = {}


class _PartyModuleProxy:
    """Lazy module proxy bound to a specific party.

    Attribute access resolves a callable inside the registered module and
    returns a wrapped function that executes with the party's mask.
    Non-callable attributes are rejected explicitly to keep semantics clear.
    """

    def __init__(self, party: Party, name: str):
        self._party: Party = party
        self._name: str = name
        self._module: ModuleType | None = None  # loaded lazily

    def _ensure(self) -> None:
        if self._module is None:
            self._module = importlib.import_module(_NAMESPACE_REGISTRY[self._name])

    def __getattr__(self, item: str) -> Callable[..., Any]:
        self._ensure()
        target = getattr(self._module, item)
        if not callable(target):
            raise AttributeError(
                f"Attribute '{item}' of party module '{self._name}' is not callable (got {type(target).__name__})"
            )

        @wraps(target)
        def _wrapped(*args: Any, **kw: Any) -> Any:
            # Inline runAt to reduce an extra partial layer while preserving semantics.
            return run_impl(Mask.from_ranks(self._party.rank), target, *args, **kw)

        # Provide a party-qualified name for debugging / logs without losing original metadata.
        base_name = getattr(target, "__name__", None)
        if base_name is None:
            # Frontend FeOperation or object without __name__; try .name attribute (FeOperation contract) or fallback to repr
            base_name = getattr(target, "name", None) or type(target).__name__
        try:
            _wrapped.__name__ = f"{base_name}@P{self._party.rank}"
        except Exception:  # pragma: no cover - assignment may fail for exotic wrappers
            pass
        return _wrapped


class Party:
    def __init__(self, rank: int) -> None:
        self.rank: int = int(rank)

    def __repr__(self) -> str:  # pragma: no cover
        return f"Party(rank={self.rank})"

    def __call__(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        if not callable(fn):
            raise TypeError(
                f"First argument to Party({self.rank}) must be callable, got {fn!r}"
            )
        return runAt(self.rank, fn)(*args, **kwargs)

    def __getattr__(self, name: str) -> _PartyModuleProxy:
        if name in _NAMESPACE_REGISTRY:
            return _PartyModuleProxy(self, name)
        raise AttributeError(
            f"Party has no attribute '{name}'. Registered: {list(_NAMESPACE_REGISTRY)}"
        )


class _PartyIndex:
    def __getitem__(self, rank: int) -> Party:
        return Party(rank)


def _load_prelude_modules() -> None:
    """Auto-register public frontend submodules for party namespace access.

    Implementation detail: we treat every non-underscore immediate child of
    ``mplang.ops`` as public and make it available as ``P0.<name>``.
    This keeps user ergonomics high (no manual load_module calls for core
    frontends) but slightly increases implicit surface area. If this grows
    unwieldy we can switch to an allowlist.
    """
    try:
        import mplang.ops as _fe  # type: ignore
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        # Frontend package not present (minimal install); safe to skip.
        return

    pkg_path = pathlib.Path(_fe.__file__).parent
    for m in pkgutil.iter_modules([str(pkg_path)]):
        if m.name.startswith("_"):
            continue
        if m.name not in _NAMESPACE_REGISTRY:
            _NAMESPACE_REGISTRY[m.name] = f"mplang.ops.{m.name}"


def load_module(module: str, alias: str | None = None) -> None:
    """Register a module for party-scoped (per-rank) callable access.

    After registration, each party object (e.g. ``P0``) can access callable
    attributes of the target module through the chosen alias and have them
    executed under that party's mask automatically. Non-callable attributes
    are intentionally not exposed to avoid ambiguity between local data and
    distributed execution semantics.

    Parameters
    ----------
    module : str
        The fully-qualified import path of the module to expose. It must be
        importable via ``importlib.import_module``.
    alias : str | None, optional
        The short name used as an attribute on ``Party``/``P0``/``P1``/... .
        If omitted, the last path segment of ``module`` is used.

    Raises
    ------
    ValueError
        If the alias is already registered to a *different* module path.

    Notes
    -----
    Registration is idempotent when the alias maps to the same module. The
    actual module object is imported lazily on first attribute access, so
    calling ``load_module`` has negligible upfront cost.

    Examples
    --------
    >>> load_module("mplang.ops.crypto", alias="crypto")
    >>> # Now call an op on party 0
    >>> P0.crypto.encrypt(data)
    """
    if alias is None:
        alias = module.rsplit(".", 1)[-1]
    prev = _NAMESPACE_REGISTRY.get(alias)
    if prev and prev != module:
        raise ValueError(f"Alias '{alias}' already registered for '{prev}'")
    _NAMESPACE_REGISTRY[alias] = module


P = _PartyIndex()
P0, P1, P2 = Party(0), Party(1), Party(2)

_load_prelude_modules()



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

import numpy as np
import pytest

from mplang.core.pfunc import PFunction
from mplang.core.tensor import TensorType
from mplang.kernels.context import RuntimeContext


class TestBuiltin:
    def setup_method(self):
        # initialize backend context for rank 0 (world_size=1) once per test
        self.runtime = RuntimeContext(rank=0, world_size=1)

    @staticmethod
    def _exec_static(runtime, pfunc: PFunction, args: list):
        return runtime.run_kernel(pfunc, args)

    def _exec(self, pfunc: PFunction, args: list):  # instance helper
        return self._exec_static(self.runtime, pfunc, args)

    def test_list_fn_names(self):
        """Test that handler lists correct function names."""
        # list_registered_kernels lives in backend.base; import lazily
        from mplang.kernels.base import list_kernels

        kernels = list_kernels()
        for expected in [
            "builtin.identity",
            "builtin.read",
            "builtin.write",
            "builtin.constant",
            "builtin.rank",
            "builtin.prand",
            "builtin.table_to_tensor",
            "builtin.tensor_to_table",
            "builtin.debug_print",
        ]:
            assert expected in kernels

    def test_identity(self):
        """Test identity operation."""
        # Create test data
        test_data = np.array([1, 2, 3], dtype=np.float32)

        # Test identity
        identity_pfunc = PFunction(
            fn_type="builtin.identity",
            ins_info=(TensorType.from_obj(test_data),),
            outs_info=(TensorType.from_obj(test_data),),
            fn_name="Identity",
        )
        result = self._exec(identity_pfunc, [test_data])
        assert len(result) == 1
        assert np.array_equal(result[0], test_data)
        assert result[0] is test_data  # Should be the same object

    def test_write_and_read_numpy_array(self):
        """Test writing and reading a numpy array."""
        # Create test data
        test_data = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)

        with tempfile.NamedTemporaryFile(suffix=".npy", delete=False) as tmp_file:
            tmp_path = tmp_file.name

        try:
            # Test write
            write_pfunc = PFunction(
                fn_type="builtin.write",
                ins_info=(TensorType.from_obj(test_data),),
                outs_info=(TensorType.from_obj(test_data),),
                fn_name="Write",
                path=tmp_path,
            )

            write_result = self._exec(write_pfunc, [test_data])
            assert len(write_result) == 1
            assert np.array_equal(write_result[0], test_data)

            # Verify file was created
            assert os.path.exists(tmp_path)

            # Test read
            read_pfunc = PFunction(
                fn_type="builtin.read",
                ins_info=(),
                outs_info=(TensorType.from_obj(test_data),),
                fn_name="Read",
                path=tmp_path,
            )

            read_result = self._exec(read_pfunc, [])
            assert len(read_result) == 1
            assert np.array_equal(read_result[0], test_data)
            assert read_result[0].dtype == test_data.dtype

        finally:
            # Clean up
            if os.path.exists(tmp_path):
                os.unlink(tmp_path)

    def test_write_creates_directory(self):
        """Test that write creates directory if it doesn't exist."""
        test_data = np.array([1, 2, 3], dtype=np.int32)

        with tempfile.TemporaryDirectory() as tmp_dir:
            nested_path = os.path.join(tmp_dir, "nested", "deep", "file.npy")

            write_pfunc = PFunction(
                fn_type="builtin.write",
                ins_info=(TensorType.from_obj(test_data),),
                outs_info=(TensorType.from_obj(test_data),),
                fn_name="Write",
                path=nested_path,
            )

            result = self._exec(write_pfunc, [test_data])
            assert len(result) == 1
            assert np.array_equal(result[0], test_data)
            assert os.path.exists(nested_path)

            # Verify the file content
            loaded = np.load(nested_path)
            assert np.array_equal(loaded, test_data)

    def test_write_different_tensor_types(self):
        """Test writing different types of tensor-like objects."""
        test_cases = [
            np.array([1, 2, 3], dtype=np.int32),
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([True, False, True], dtype=bool),
            42,  # scalar
            np.array([1, 2, 3]),  # converted list to numpy array
        ]

        for i, test_data in enumerate(test_cases):
            with tempfile.NamedTemporaryFile(
                suffix=f"_{i}.npy", delete=False
            ) as tmp_file:
                tmp_path = tmp_file.name

            try:
                tensor_info = TensorType.from_obj(test_data)
                write_pfunc = PFunction(
                    fn_type="builtin.write",
                    ins_info=(tensor_info,),
                    outs_info=(tensor_info,),
                    fn_name="Write",
                    path=tmp_path,
                )

                result = self._exec(write_pfunc, [test_data])
                assert len(result) == 1

                # Read back and verify
                read_pfunc = PFunction(
                    fn_type="builtin.read",
                    ins_info=(),
                    outs_info=(tensor_info,),
                    fn_name="Read",
                    path=tmp_path,
                )

                read_result = self._exec(read_pfunc, [])
                assert len(read_result) == 1
                assert np.array_equal(read_result[0], np.array(test_data))

            finally:
                if os.path.exists(tmp_path):
                    os.unlink(tmp_path)

    def test_identity_wrong_args(self):
        """Test identity with wrong number of arguments."""
        # Declare expected single input so runtime enforces arity directly.
        identity_pfunc = PFunction(
            fn_type="builtin.identity",
            ins_info=(TensorType.from_obj(np.array(1, dtype=np.int32)),),
            outs_info=(TensorType.from_obj(np.array(1, dtype=np.int32)),),
            fn_name="Identity",
        )
        with pytest.raises(ValueError, match=r"arg count mismatch: got 0, expect 1"):
            self._exec(identity_pfunc, [])
        with pytest.raises(ValueError, match=r"arg count mismatch: got 2, expect 1"):
            self._exec(identity_pfunc, [1, 2])

    def test_read_missing_path(self):
        """Test read operation without path attribute."""
        read_pfunc = PFunction(
            fn_type="builtin.read",
            ins_info=(),
            outs_info=(TensorType.from_obj(np.array([1])),),
            fn_name="Read",
        )
        with pytest.raises(ValueError, match=r"missing path attr for builtin.read"):
            self._exec(read_pfunc, [])

    def test_read_wrong_args(self):
        """Test read with wrong number of arguments."""
        read_pfunc = PFunction(
            fn_type="builtin.read",
            ins_info=(),
            outs_info=(TensorType.from_obj(np.array([1])),),
            fn_name="Read",
            path="dummy.npy",
        )
        with pytest.raises(ValueError, match=r"arg count mismatch: got 1, expect 0"):
            self._exec(read_pfunc, [np.array([1])])

    def test_write_missing_path(self):
        """Test write operation without path attribute."""
        test_data = np.array([1, 2, 3])
        write_pfunc = PFunction(
            fn_type="builtin.write",
            ins_info=(TensorType.from_obj(test_data),),
            outs_info=(TensorType.from_obj(test_data),),
            fn_name="Write",
        )
        with pytest.raises(ValueError, match=r"missing path attr for builtin.write"):
            self._exec(write_pfunc, [test_data])

    def test_write_wrong_args(self):
        """Test write with wrong number of arguments."""
        write_pfunc = PFunction(
            fn_type="builtin.write",
            ins_info=(TensorType.from_obj(np.array(1, dtype=np.int32)),),
            outs_info=(TensorType.from_obj(np.array(1, dtype=np.int32)),),
            fn_name="Write",
            path="dummy.npy",
        )
        with pytest.raises(ValueError, match=r"arg count mismatch: got 0, expect 1"):
            self._exec(write_pfunc, [])
        with pytest.raises(ValueError, match=r"arg count mismatch: got 2, expect 1"):
            self._exec(write_pfunc, [1, 2])

    def test_read_nonexistent_file(self):
        """Test reading from a nonexistent file."""
        read_pfunc = PFunction(
            fn_type="builtin.read",
            ins_info=(),
            outs_info=(TensorType.from_obj(np.array([1])),),
            fn_name="Read",
            path="nonexistent_file.npy",
        )
        with pytest.raises(RuntimeError, match=r"builtin.read failed:"):
            self._exec(read_pfunc, [])

    def test_unsupported_function_type(self):
        """Test unsupported function type."""
        pfunc = PFunction(
            fn_type="unsupported.operation",
            ins_info=(),
            outs_info=(),
            fn_name="Unsupported",
        )

        with pytest.raises(NotImplementedError, match="no backend kernel registered"):
            self._exec(pfunc, [])



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
HTTP-based driver implementation for distributed execution.

This module provides an HTTP-based alternative to the gRPC Driver,
using REST APIs for distributed multi-party computation coordination.
"""

from __future__ import annotations

import asyncio
import base64
import uuid
from collections.abc import Sequence
from typing import Any

from mplang.core.cluster import ClusterSpec
from mplang.core.expr.ast import Expr
from mplang.core.interp import InterpContext, InterpVar
from mplang.core.mask import Mask
from mplang.core.mpir import Writer
from mplang.core.mpobject import MPObject
from mplang.core.mptype import MPType
from mplang.runtime.client import HttpExecutorClient


def new_uuid() -> str:
    """Generates a short UUID using URL-safe Base64 encoding."""
    u = uuid.uuid4()
    # Get the 16 bytes of the UUID
    uuid_bytes = u.bytes
    # Encode using URL-safe Base64
    encoded_bytes = base64.urlsafe_b64encode(uuid_bytes)
    # Decode to UTF-8 string, remove padding, and take first 8 characters
    encoded_string = encoded_bytes.decode("utf-8").rstrip("=")[:8]
    return encoded_string


class DriverVar(InterpVar):
    """A variable that references a value in distributed HTTP executor nodes.

    This represents a symbol stored on remote HTTP servers that can be
    retrieved via REST API calls.
    """

    def __init__(
        self,
        ctx: Driver,
        symbol_name: str,
        mptype: MPType,
    ) -> None:
        super().__init__(ctx, mptype)
        self.symbol_name = symbol_name

    @property
    def mptype(self) -> MPType:
        """The type of this variable."""
        return self._mptype

    def __repr__(self) -> str:
        return f"HttpDriverVar(symbol_name={self.symbol_name}, mptype={self.mptype})"


class Driver(InterpContext):
    """Driver for distributed execution using HTTP-based services.

    Args:
        cluster_spec: The cluster specification defining the distributed environment.
        trace_ranks: List of ranks to trace execution for debugging.
        timeout: HTTP request timeout in seconds.
    """

    def __init__(
        self,
        cluster_spec: ClusterSpec,
        *,
        trace_ranks: list[int] | None = None,
        timeout: int = 120,
    ) -> None:
        """Initialize a driver with the given cluster specification.

        Args:
            cluster_spec: The cluster specification defining the distributed environment.
            trace_ranks: List of ranks to trace execution for debugging.
            timeout: HTTP request timeout in seconds.
        """
        super().__init__(cluster_spec)
        self._trace_ranks = trace_ranks or []
        self.timeout = timeout

        self._session_id: str | None = None
        self._counter = 0

        self.node_addrs = {
            node_id: node.endpoint for node_id, node in cluster_spec.nodes.items()
        }

        # Get SPU configuration from cluster_spec
        spu_devices = cluster_spec.get_devices_by_kind("SPU")
        if not spu_devices:
            raise ValueError("No SPU device found in the cluster specification")
        if len(spu_devices) > 1:
            raise ValueError("Multiple SPU devices found in the cluster specification")
        spu_device = spu_devices[0]

        # Store SPU configuration as strings for better readability
        self.spu_protocol_str = spu_device.config["protocol"]
        self.spu_field_str = spu_device.config["field"]

        # Compute spu_mask from spu_device members
        spu_mask = Mask.from_ranks([member.rank for member in spu_device.members])
        self.spu_mask_int = int(spu_mask)

    def _create_clients(self) -> dict[str, HttpExecutorClient]:
        """Create HTTP clients for all endpoints."""
        clients = {}
        for node_id, endpoint in self.node_addrs.items():
            clients[node_id] = HttpExecutorClient(endpoint, self.timeout)
        return clients

    async def _close_clients(self, clients: dict[str, HttpExecutorClient]) -> None:
        """Close all provided HTTP clients."""
        await asyncio.gather(*[client.close() for client in clients.values()])

    def new_name(self, prefix: str = "var") -> str:
        """Generate a unique execution name."""
        name = f"{prefix}_{self._counter}"
        self._counter += 1
        return name

    async def _get_or_create_session(self) -> str:
        """Get existing session or create a new one across all HTTP servers."""
        if self._session_id is None:
            new_session_id = new_uuid()
            # Create temporary clients for session creation
            clients = self._create_clients()
            try:
                # Create session on all HTTP servers concurrently
                tasks = []
                for node_id, client in clients.items():
                    # Convert node_id to rank for the session creation
                    rank = list(self.node_addrs.keys()).index(node_id)
                    task = client.create_session(
                        name=new_session_id,
                        rank=rank,
                        cluster_spec=self.cluster_spec.to_dict(),
                    )
                    tasks.append(task)

                try:
                    results = await asyncio.gather(*tasks)
                    for session_id in results:
                        assert session_id == new_session_id
                    self._session_id = new_session_id
                except RuntimeError as e:
                    raise RuntimeError(
                        f"Failed to create session on one or more parties: {e}"
                    ) from e
            finally:
                await self._close_clients(clients)

        assert self._session_id is not None
        return self._session_id

    async def _evaluate(
        self, expr: Expr, bindings: dict[str, MPObject]
    ) -> Sequence[MPObject]:
        """Async implementation to evaluate an expression."""
        session_id = await self._get_or_create_session()

        # Prepare input names from bindings
        var_names = []
        party_symbol_names = []
        for name, var in bindings.items():
            if var.ctx is not self:
                raise ValueError(f"Variable {name} not in this context, got {var.ctx}.")
            assert isinstance(var, DriverVar), (
                f"Expected HttpDriverVar, got {type(var)}"
            )
            var_names.append(name)
            party_symbol_names.append(var.symbol_name)

        var_name_mapping = dict(zip(var_names, party_symbol_names, strict=True))

        writer = Writer(var_name_mapping)
        program_proto = writer.dumps(expr)

        output_symbols = [self.new_name() for _ in range(expr.num_outputs)]

        # Create temporary clients for computation execution
        clients = self._create_clients()
        try:
            # Concurrently create and execute computation on all parties
            tasks = []
            computation_id = new_uuid()
            for _rank, client in clients.items():
                task = client.create_and_execute_computation(
                    session_id,
                    computation_id,
                    program_proto.SerializeToString(),
                    party_symbol_names,
                    output_symbols,
                )
                tasks.append(task)

            try:
                await asyncio.gather(*tasks)
            except RuntimeError as e:
                raise RuntimeError(
                    f"Failed to create and execute computation on one or more parties: {e}"
                ) from e
        finally:
            await self._close_clients(clients)

        # Create HttpDriverVar objects for each output
        driver_vars = []
        for symbol_name, mptype in zip(output_symbols, expr.mptypes, strict=True):
            driver_var = DriverVar(self, symbol_name, mptype)
            driver_vars.append(driver_var)

        return driver_vars

    def evaluate(self, expr: Expr, bindings: dict[str, MPObject]) -> Sequence[MPObject]:
        """Evaluate an expression using distributed HTTP execution."""
        return asyncio.run(self._evaluate(expr, bindings))

    async def _fetch(self, obj: MPObject) -> list[Any]:
        """Async implementation to fetch results."""
        if not isinstance(obj, DriverVar):
            raise ValueError(f"Expected HttpDriverVar, got {type(obj)}")

        session_id = await self._get_or_create_session()
        symbol_full_name = obj.symbol_name

        # Create temporary clients for fetching
        clients = self._create_clients()
        try:
            # Concurrently fetch symbol from all parties
            tasks = []
            for _rank, client in clients.items():
                task = client.get_symbol(session_id, symbol_full_name)
                tasks.append(task)

            try:
                # The results will be in the same order as the clients (ranks)
                results = await asyncio.gather(*tasks)
                return list(results)
            except RuntimeError as e:
                raise RuntimeError(
                    f"Failed to fetch symbol from one or more parties: {e}"
                ) from e
        finally:
            await self._close_clients(clients)

    def fetch(self, obj: MPObject) -> list[Any]:
        """Fetch results from the distributed HTTP execution."""
        return asyncio.run(self._fetch(obj))

    async def _ping(self, node_id: str) -> bool:
        """Async implementation to ping a node.

        Args:
            node_id: The ID of the node to ping

        Returns:
            True if the node is healthy, False otherwise
        """
        # Create a temporary client for the node
        if node_id not in self.node_addrs:
            raise ValueError(f"Node {node_id} not found in party addresses")

        endpoint = self.node_addrs[node_id]
        client = HttpExecutorClient(endpoint, self.timeout)

        try:
            # Perform health check
            return await client.health_check()
        except Exception:
            # Any exception means the node is not healthy
            return False
        finally:
            await client.close()

    def ping(self, node_id: str) -> bool:
        """Ping a node to check if it's healthy.

        Args:
            node_id: The ID of the node to ping

        Returns:
            True if the node is healthy, False otherwise
        """
        return asyncio.run(self._ping(node_id))



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core Session model (pure, no global registries).

Contents:
    * SessionState dataclass
    * LinkCommFactory (SPU link reuse cache)
    * Session (topology derivation, runtime init, SPU env seeding, local symbol/computation storage)

Process-wide registries (sessions, global symbols) live in the server layer
(`server.py`) so this module remains portable and easy to unit test.
"""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Any, cast
from urllib.parse import urlparse

import spu.libspu as libspu

from mplang.core.expr.ast import Expr
from mplang.core.expr.evaluator import IEvaluator, create_evaluator
from mplang.core.mask import Mask
from mplang.kernels.context import RuntimeContext
from mplang.kernels.spu import PFunction  # type: ignore
from mplang.runtime.communicator import HttpCommunicator
from mplang.runtime.exceptions import ResourceNotFound
from mplang.runtime.link_comm import LinkCommunicator
from mplang.utils.spu_utils import parse_field, parse_protocol

if TYPE_CHECKING:  # pragma: no cover - import only for type checking
    from mplang.core.cluster import ClusterSpec, Node, RuntimeInfo


class LinkCommFactory:
    """Factory for creating and caching link communicators."""

    def __init__(self) -> None:
        self._cache: dict[tuple[int, tuple[str, ...]], LinkCommunicator] = {}

    def create_link(self, rel_rank: int, addrs: list[str]) -> LinkCommunicator:
        key = (rel_rank, tuple(addrs))
        link = self._cache.get(key)
        if link is not None:
            return link
        logging.info(f"LinkCommunicator created: rel_rank={rel_rank} addrs={addrs}")
        link = LinkCommunicator(rel_rank, addrs)
        self._cache[key] = link
        return link


# Shared link factory (module-local, not global registry of sessions)
g_link_factory = LinkCommFactory()


@dataclass
class Symbol:
    name: str
    mptype: Any
    data: Any


@dataclass
class Computation:
    name: str
    expr: Expr


@dataclass
class SessionState:
    runtime: RuntimeContext | None = None
    computations: dict[str, Computation] = field(default_factory=dict)
    symbols: dict[str, Symbol] = field(default_factory=dict)
    spu_seeded: bool = False
    created_ts: float = field(default_factory=time.time)
    last_access_ts: float = field(default_factory=time.time)


class Session:
    """Represents the per-rank execution context.

    Immutable config: name, rank, cluster_spec.
    Derived: node, runtime_info, endpoints, spu_device, spu_mask, protocol/field, is_spu_party.
    Mutable: state (runtime object, symbols, computations, seeded flag).
    """

    def __init__(self, name: str, rank: int, cluster_spec: ClusterSpec):
        self.name = name
        self.rank = rank
        self.cluster_spec = cluster_spec
        self.state = SessionState()
        self.communicator = HttpCommunicator(
            session_name=name, rank=rank, endpoints=self.endpoints
        )

    # --- Derived topology ---
    @cached_property
    def node(self) -> Node:
        return self.cluster_spec.get_node_by_rank(self.rank)

    @property
    def runtime_info(self) -> RuntimeInfo:
        return self.node.runtime_info

    @cached_property
    def endpoints(self) -> list[str]:
        eps: list[str] = []
        for n in sorted(
            self.cluster_spec.nodes.values(),
            key=lambda x: x.rank,  # type: ignore[attr-defined]
        ):
            ep = n.endpoint
            if not ep.startswith(("http://", "https://")):
                ep = f"http://{ep}"
            eps.append(ep)
        return eps

    @cached_property
    def spu_device(self):  # type: ignore
        devs = self.cluster_spec.get_devices_by_kind("SPU")
        if len(devs) != 1:
            raise RuntimeError(
                f"Expected exactly one SPU device, got {len(devs)} (session={self.name})"
            )
        return devs[0]

    @cached_property
    def spu_mask(self) -> Mask:
        return Mask.from_ranks([m.rank for m in self.spu_device.members])

    @property
    def spu_protocol(self) -> str:
        return cast(str, self.spu_device.config.get("protocol", "SEMI2K"))

    @property
    def spu_field(self) -> str:
        return cast(str, self.spu_device.config.get("field", "FM64"))

    @property
    def is_spu_party(self) -> bool:
        return self.rank in self.spu_mask

    # --- Runtime helpers ---
    def ensure_runtime(self) -> RuntimeContext:
        if self.state.runtime is None:
            self.state.runtime = RuntimeContext(
                rank=self.rank,
                world_size=len(self.cluster_spec.nodes),  # type: ignore[attr-defined]
                initial_bindings=(
                    self.runtime_info.op_bindings if self.runtime_info else {}
                ),
            )
        return self.state.runtime

    def ensure_spu_env(self) -> None:
        """Ensure SPU kernel env (config/world[/link]) registered on this runtime.

        Previous logic only seeded SPU parties; non-participating ranks then raised
        a hard error when the evaluator encountered SPU ops in the global program,
        because the kernel pocket lacked config/world. For now we register the
        config/world on ALL parties (idempotent) and only attach a link context for
        participating SPU ranks. Non-parties will still error later if they try to
        execute a link-dependent SPU kernel (which should be guarded by masks in the
        IR), but they will no longer fail early with a misleading
        "SPU kernel state not initialized" message.
        """
        if self.state.spu_seeded:
            return

        link_ctx = None
        # TODO(jint): reuse same port for mplang and spu.
        SPU_PORT_OFFSET = 100

        if self.is_spu_party:
            # Build SPU address list across all endpoints for ranks in mask
            spu_addrs: list[str] = []
            for r, addr in enumerate(self.communicator.endpoints):
                if r in self.spu_mask:
                    if "//" not in addr:
                        addr = f"//{addr}"
                    parsed = urlparse(addr)
                    assert isinstance(parsed.port, int)
                    new_addr = f"{parsed.hostname}:{parsed.port + SPU_PORT_OFFSET}"
                    spu_addrs.append(new_addr)
            rel_index = sum(1 for r in range(self.rank) if r in self.spu_mask)
            link_ctx = g_link_factory.create_link(rel_index, spu_addrs)

        spu_config = libspu.RuntimeConfig(
            protocol=parse_protocol(self.spu_protocol),
            field=parse_field(self.spu_field),
            fxp_fraction_bits=18,
        )
        seed_pfunc = PFunction(
            fn_type="spu.seed_env",
            ins_info=(),
            outs_info=(),
            config=spu_config,
            world=self.spu_mask.num_parties(),
            link=link_ctx,
        )
        self.ensure_runtime().run_kernel(seed_pfunc, [])
        self.state.spu_seeded = True

    # --- Computations & Symbols (instance-local) ---
    def add_computation(self, computation: Computation) -> None:
        self.state.computations[computation.name] = computation

    def get_computation(self, name: str) -> Computation | None:
        return self.state.computations.get(name)

    def add_symbol(self, symbol: Symbol) -> None:
        self.state.symbols[symbol.name] = symbol

    def get_symbol(self, name: str) -> Symbol | None:
        return self.state.symbols.get(name)

    def list_symbols(self) -> list[str]:  # pragma: no cover - trivial
        return list(self.state.symbols.keys())

    def delete_symbol(self, name: str) -> bool:
        if name in self.state.symbols:
            del self.state.symbols[name]
            return True
        return False

    def list_computations(self) -> list[str]:  # pragma: no cover - trivial
        return list(self.state.computations.keys())

    def delete_computation(self, name: str) -> bool:
        if name in self.state.computations:
            del self.state.computations[name]
            return True
        return False

    # --- Execution ---
    def execute(
        self, computation: Computation, input_names: list[str], output_names: list[str]
    ) -> None:
        env: dict[str, Any] = {}
        for in_name in input_names:
            sym = self.get_symbol(in_name)
            if sym is None:
                raise ResourceNotFound(
                    f"Input symbol '{in_name}' not found in session '{self.name}'"
                )
            env[in_name] = sym.data
        rt = self.ensure_runtime()
        self.ensure_spu_env()
        evaluator: IEvaluator = create_evaluator(
            rank=self.rank, env=env, comm=self.communicator, runtime=rt
        )
        results = evaluator.evaluate(computation.expr)
        if results and len(results) != len(output_names):
            raise RuntimeError(
                f"Expected {len(output_names)} results, got {len(results)}"
            )
        for name, val in zip(output_names, results, strict=True):
            self.add_symbol(Symbol(name=name, mptype={}, data=val))

    # --- Convenience constructor ---
    @classmethod
    def from_cluster_spec_dict(cls, name: str, rank: int, spec_dict: dict) -> Session:
        from mplang.core.cluster import ClusterSpec  # local import to avoid cycles

        spec = ClusterSpec.from_dict(spec_dict)
        if len(spec.get_devices_by_kind("SPU")) == 0:
            raise RuntimeError("No SPU device found in cluster_spec")
        return cls(name=name, rank=rank, cluster_spec=spec)



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

import mplang
import mplang.simp as simp
import mplang.simp.random as mpr


def eval_and_fetch(sim, fn, *args, **kwargs):
    """Helper function to evaluate a function and fetch results."""
    result = mplang.evaluate(sim, fn, *args, **kwargs)
    return mplang.fetch(sim, result)


class TestPShfl:
    """Test pshfl related functions"""

    def test_pshfl_basic(self):
        """Test basic pshfl_s functionality"""
        num_parties = 10
        sim = mplang.Simulator.simple(num_parties)

        src = mplang.evaluate(sim, mpr.prandint, 0, 100)
        key = mplang.evaluate(sim, mpr.ukey, 42)
        index = mplang.evaluate(sim, mpr.pperm, key)

        # shuffle data with range, nothing changed.
        shuffled = mplang.evaluate(sim, simp.pshfl, src, index)

        data, index, shuffled = mplang.fetch(sim, (src, index, shuffled))
        data, index, shuffled = np.stack(data), np.stack(index), np.stack(shuffled)
        np.testing.assert_array_equal(data[index], shuffled)

    # TODO(jint): add shfl complicated


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
from collections.abc import Sequence
from types import MappingProxyType
from typing import Any

from mplang.core.table import TableType
from mplang.core.tensor import TensorType

__all__ = [
    "PFunction",
    "get_fn_name",
]


class PFunction:
    """A Party Function represents a computation unit that can be executed by a single party.

    PFunction serves as a unified interface for describing single-party computations
    in multi-party computing scenarios. It can represent both:
    1. Built-in operations (e.g., "spu.makeshares", "builtin.read")
    2. User-defined programmable functions with custom code

    The PFunction accepts a list of typed inputs (TensorType/TableType). For
    backend-only handles (e.g., crypto keys), use a sentinel TensorType
    of UINT8 with shape (-1, 0) to indicate the argument should bypass
    structural validation at runtime. Outputs should likewise use concrete
    TensorType/TableType specs. PFunction can be:
    - Expressed and defined in the mplang frontend
    - Serialized for transmission between components
    - Interpreted and executed by backend runtime engines

    Args:
        fn_type: The type/category identifier of this PFunction, indicating which
            backend or handler should process it (e.g., "spu.makeshares", "builtin.read",
            "mlir.stablehlo"). This serves as a routing mechanism for execution.
        ins_info: Type information for input parameters (TensorType or TableType)
        outs_info: Type information for output values (TensorType or TableType)
        fn_name: Optional name of the function. For programmable functions, this is
            the user-defined function name. For built-in operations, this may be
            None or a descriptive identifier.
        fn_text: Optional serialized function body. For programmable functions, this
            contains the actual code (e.g., MLIR, bytecode, source code). For built-in
            operations, this is typically None.
        **kwargs: Additional attributes and metadata specific to the function type.
            These are used to pass execution parameters, configuration, and context
            information to the backend handlers.
    """

    # Required fields - these define the core execution context
    fn_type: str  # Unique identifier for backend routing
    ins_info: tuple[TensorType | TableType, ...]
    outs_info: tuple[TensorType | TableType, ...]

    # Optional fields for programmable functions
    fn_name: str | None  # Function name (for programmable functions)
    fn_text: str | None  # Function body/code (for programmable functions)

    # Custom attributes and metadata
    attrs: MappingProxyType[str, Any]  # Execution parameters and metadata

    def __init__(
        self,
        fn_type: str,
        ins_info: Sequence[TensorType | TableType],
        outs_info: Sequence[TensorType | TableType],
        *,
        fn_name: str | None = None,
        fn_text: str | None = None,
        **kwargs: Any,
    ):
        self.fn_type = fn_type
        self.fn_name = fn_name
        self.fn_text = fn_text
        self.ins_info = tuple(ins_info)
        self.outs_info = tuple(outs_info)
        # Make attrs immutable to ensure PFunction immutability
        # Create a copy first, then wrap it in MappingProxyType
        self.attrs = MappingProxyType(copy.copy(kwargs))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.fn_type}, {self.fn_name})"

    def __hash__(self) -> int:
        return hash((
            self.fn_type,
            self.fn_name,
            self.fn_text,
            self.ins_info,
            self.outs_info,
            frozenset(self.attrs.items()),
        ))

    def __eq__(self, other: object) -> bool:
        """Check equality between PFunction instances."""
        if not isinstance(other, PFunction):
            return False

        return (
            self.fn_type == other.fn_type
            and self.fn_name == other.fn_name
            and self.fn_text == other.fn_text
            and self.ins_info == other.ins_info
            and self.outs_info == other.outs_info
            and self.attrs == other.attrs
        )


def get_fn_name(fn_like: Any) -> str:
    if hasattr(fn_like, "__name__"):
        return fn_like.__name__  # type: ignore[no-any-return]
    if hasattr(fn_like, "func"):
        # handle partial functions
        return get_fn_name(fn_like.func)
    return "unnamed function"



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax.tree_util import PyTreeDef, tree_flatten

from mplang.core.mpobject import MPObject
from mplang.core.pfunc import PFunction
from mplang.core.table import TableType
from mplang.ops.base import FeOperation, stateless_mod

_SQL_MOD = stateless_mod("sql")


class SqlFE(FeOperation):
    def __init__(self, dialect: str = "duckdb"):
        # Bind to sql module with a stable op name for registry/dispatch
        super().__init__(_SQL_MOD, "run")
        self._dialect = dialect

    def trace(
        self,
        sql: str,
        out_type: TableType,
        in_tables: dict[str, MPObject] | None = None,
    ) -> tuple[PFunction, list[MPObject], PyTreeDef]:
        in_names: list[str] = []
        ins_info: list[TableType] = []
        in_vars: list[MPObject] = []
        if in_tables:
            for name, tbl in in_tables.items():
                assert isinstance(tbl, MPObject)
                assert tbl.schema is not None
                in_names.append(name)
                ins_info.append(tbl.schema)
                in_vars.append(tbl)

        pfn = PFunction(
            fn_type="sql.run",
            fn_name="",
            fn_text=sql,
            ins_info=tuple(ins_info),
            outs_info=(out_type,),
            in_names=tuple(in_names),
            dialect=self._dialect,
        )
        _, treedef = tree_flatten(out_type)
        return pfn, in_vars, treedef


sql_run = SqlFE("duckdb")



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2025 Ant Group Co., Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _DataType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DataTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DataType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNDEFINED: _DataType.ValueType  # 0
    U8: _DataType.ValueType  # 1
    """uint8_t"""
    I8: _DataType.ValueType  # 2
    """int8_t"""
    U16: _DataType.ValueType  # 3
    """uint16_t"""
    I16: _DataType.ValueType  # 4
    """int16_t"""
    U32: _DataType.ValueType  # 5
    """uint32_t"""
    I32: _DataType.ValueType  # 6
    """int32_t"""
    U64: _DataType.ValueType  # 7
    """uint64_t"""
    I64: _DataType.ValueType  # 8
    """int64_t"""
    F16: _DataType.ValueType  # 9
    """float16"""
    F32: _DataType.ValueType  # 10
    """float32"""
    F64: _DataType.ValueType  # 11
    """double"""
    STRING: _DataType.ValueType  # 12
    """string (table only)"""
    BOOL: _DataType.ValueType  # 13
    """bool"""
    COMPLEX64: _DataType.ValueType  # 14
    """complex with float32 real and imaginary components"""
    COMPLEX128: _DataType.ValueType  # 15
    """complex with float64 real and imaginary components"""
    BFLOAT16: _DataType.ValueType  # 16
    """bfloat16"""
    DATE: _DataType.ValueType  # 17
    """date (table only)"""
    TIME: _DataType.ValueType  # 18
    """time (table only)"""
    TIMESTAMP: _DataType.ValueType  # 19
    """timestamp (table only)"""
    JSON: _DataType.ValueType  # 20
    """json (table only)"""

class DataType(_DataType, metaclass=_DataTypeEnumTypeWrapper):
    """Data type enumeration"""

UNDEFINED: DataType.ValueType  # 0
U8: DataType.ValueType  # 1
"""uint8_t"""
I8: DataType.ValueType  # 2
"""int8_t"""
U16: DataType.ValueType  # 3
"""uint16_t"""
I16: DataType.ValueType  # 4
"""int16_t"""
U32: DataType.ValueType  # 5
"""uint32_t"""
I32: DataType.ValueType  # 6
"""int32_t"""
U64: DataType.ValueType  # 7
"""uint64_t"""
I64: DataType.ValueType  # 8
"""int64_t"""
F16: DataType.ValueType  # 9
"""float16"""
F32: DataType.ValueType  # 10
"""float32"""
F64: DataType.ValueType  # 11
"""double"""
STRING: DataType.ValueType  # 12
"""string (table only)"""
BOOL: DataType.ValueType  # 13
"""bool"""
COMPLEX64: DataType.ValueType  # 14
"""complex with float32 real and imaginary components"""
COMPLEX128: DataType.ValueType  # 15
"""complex with float64 real and imaginary components"""
BFLOAT16: DataType.ValueType  # 16
"""bfloat16"""
DATE: DataType.ValueType  # 17
"""date (table only)"""
TIME: DataType.ValueType  # 18
"""time (table only)"""
TIMESTAMP: DataType.ValueType  # 19
"""timestamp (table only)"""
JSON: DataType.ValueType  # 20
"""json (table only)"""
global___DataType = DataType

@typing_extensions.final
class AttrProto(google.protobuf.message.Message):
    """Custom attribute message (without name field since it's the map key)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _AttrType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _AttrTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[AttrProto._AttrType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNDEFINED: AttrProto._AttrType.ValueType  # 0
        FLOAT: AttrProto._AttrType.ValueType  # 1
        INT: AttrProto._AttrType.ValueType  # 2
        STRING: AttrProto._AttrType.ValueType  # 3
        BOOL: AttrProto._AttrType.ValueType  # 4
        BYTES: AttrProto._AttrType.ValueType  # 5
        FLOATS: AttrProto._AttrType.ValueType  # 6
        INTS: AttrProto._AttrType.ValueType  # 7
        STRINGS: AttrProto._AttrType.ValueType  # 8
        FUNCTION: AttrProto._AttrType.ValueType  # 10
        """Textual function reference"""
        GRAPH: AttrProto._AttrType.ValueType  # 11
        """Nested function graph"""

    class AttrType(_AttrType, metaclass=_AttrTypeEnumTypeWrapper):
        """Define possible attribute types"""

    UNDEFINED: AttrProto.AttrType.ValueType  # 0
    FLOAT: AttrProto.AttrType.ValueType  # 1
    INT: AttrProto.AttrType.ValueType  # 2
    STRING: AttrProto.AttrType.ValueType  # 3
    BOOL: AttrProto.AttrType.ValueType  # 4
    BYTES: AttrProto.AttrType.ValueType  # 5
    FLOATS: AttrProto.AttrType.ValueType  # 6
    INTS: AttrProto.AttrType.ValueType  # 7
    STRINGS: AttrProto.AttrType.ValueType  # 8
    FUNCTION: AttrProto.AttrType.ValueType  # 10
    """Textual function reference"""
    GRAPH: AttrProto.AttrType.ValueType  # 11
    """Nested function graph"""

    TYPE_FIELD_NUMBER: builtins.int
    F_FIELD_NUMBER: builtins.int
    I_FIELD_NUMBER: builtins.int
    S_FIELD_NUMBER: builtins.int
    B_FIELD_NUMBER: builtins.int
    RAW_BYTES_FIELD_NUMBER: builtins.int
    FLOATS_FIELD_NUMBER: builtins.int
    INTS_FIELD_NUMBER: builtins.int
    STRS_FIELD_NUMBER: builtins.int
    FUNC_FIELD_NUMBER: builtins.int
    GRAPH_FIELD_NUMBER: builtins.int
    type: global___AttrProto.AttrType.ValueType
    """Type of the attribute"""
    f: builtins.float
    """FLOAT"""
    i: builtins.int
    """INT"""
    s: builtins.str
    """STRING"""
    b: builtins.bool
    """BOOL"""
    raw_bytes: builtins.bytes
    """BYTES - for raw binary data"""
    @property
    def floats(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """FLOATS"""
    @property
    def ints(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """INTS"""
    @property
    def strs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """STRINGS"""
    @property
    def func(self) -> global___FuncProto:
        """FUNCTION - textual function reference"""
    @property
    def graph(self) -> global___GraphProto:
        """GRAPH - nested function graph"""
    def __init__(
        self,
        *,
        type: global___AttrProto.AttrType.ValueType = ...,
        f: builtins.float = ...,
        i: builtins.int = ...,
        s: builtins.str = ...,
        b: builtins.bool = ...,
        raw_bytes: builtins.bytes = ...,
        floats: collections.abc.Iterable[builtins.float] | None = ...,
        ints: collections.abc.Iterable[builtins.int] | None = ...,
        strs: collections.abc.Iterable[builtins.str] | None = ...,
        func: global___FuncProto | None = ...,
        graph: global___GraphProto | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["func", b"func", "graph", b"graph"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["b", b"b", "f", b"f", "floats", b"floats", "func", b"func", "graph", b"graph", "i", b"i", "ints", b"ints", "raw_bytes", b"raw_bytes", "s", b"s", "strs", b"strs", "type", b"type"]) -> None: ...

global___AttrProto = AttrProto

@typing_extensions.final
class FuncProto(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class AttrsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___AttrProto: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___AttrProto | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    TYPE_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    BODY_FIELD_NUMBER: builtins.int
    DOC_STRING_FIELD_NUMBER: builtins.int
    ATTRS_FIELD_NUMBER: builtins.int
    type: builtins.str
    """Function type."""
    name: builtins.str
    """Function name"""
    body: builtins.str
    """Function body as a string (e.g., Python code)"""
    doc_string: builtins.str
    """Documentation string"""
    @property
    def attrs(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___AttrProto]:
        """Function attributes"""
    def __init__(
        self,
        *,
        type: builtins.str = ...,
        name: builtins.str = ...,
        body: builtins.str = ...,
        doc_string: builtins.str = ...,
        attrs: collections.abc.Mapping[builtins.str, global___AttrProto] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attrs", b"attrs", "body", b"body", "doc_string", b"doc_string", "name", b"name", "type", b"type"]) -> None: ...

global___FuncProto = FuncProto

@typing_extensions.final
class TensorTypeProto(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DTYPE_FIELD_NUMBER: builtins.int
    SHAPE_DIMS_FIELD_NUMBER: builtins.int
    dtype: global___DataType.ValueType
    @property
    def shape_dims(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        dtype: global___DataType.ValueType = ...,
        shape_dims: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["dtype", b"dtype", "shape_dims", b"shape_dims"]) -> None: ...

global___TensorTypeProto = TensorTypeProto

@typing_extensions.final
class TableTypeProto(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class Column(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        DTYPE_FIELD_NUMBER: builtins.int
        name: builtins.str
        dtype: global___DataType.ValueType
        def __init__(
            self,
            *,
            name: builtins.str = ...,
            dtype: global___DataType.ValueType = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["dtype", b"dtype", "name", b"name"]) -> None: ...

    COLUMNS_FIELD_NUMBER: builtins.int
    @property
    def columns(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TableTypeProto.Column]: ...
    def __init__(
        self,
        *,
        columns: collections.abc.Iterable[global___TableTypeProto.Column] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["columns", b"columns"]) -> None: ...

global___TableTypeProto = TableTypeProto

@typing_extensions.final
class MPTypeProto(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class AttrsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___AttrProto: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___AttrProto | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    TENSOR_TYPE_FIELD_NUMBER: builtins.int
    TABLE_TYPE_FIELD_NUMBER: builtins.int
    PMASK_FIELD_NUMBER: builtins.int
    ATTRS_FIELD_NUMBER: builtins.int
    @property
    def tensor_type(self) -> global___TensorTypeProto: ...
    @property
    def table_type(self) -> global___TableTypeProto: ...
    pmask: builtins.int
    """party mask (-1 for dynamic mask, >=0 for static mask)"""
    @property
    def attrs(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___AttrProto]:
        """attributes"""
    def __init__(
        self,
        *,
        tensor_type: global___TensorTypeProto | None = ...,
        table_type: global___TableTypeProto | None = ...,
        pmask: builtins.int = ...,
        attrs: collections.abc.Mapping[builtins.str, global___AttrProto] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["table_type", b"table_type", "tensor_type", b"tensor_type", "type_spec", b"type_spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["attrs", b"attrs", "pmask", b"pmask", "table_type", b"table_type", "tensor_type", b"tensor_type", "type_spec", b"type_spec"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["type_spec", b"type_spec"]) -> typing_extensions.Literal["tensor_type", "table_type"] | None: ...

global___MPTypeProto = MPTypeProto

@typing_extensions.final
class NodeProto(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class AttrsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___AttrProto: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___AttrProto | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    OP_TYPE_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTS_INFO_FIELD_NUMBER: builtins.int
    ATTRS_FIELD_NUMBER: builtins.int
    DOC_STRING_FIELD_NUMBER: builtins.int
    op_type: builtins.str
    name: builtins.str
    @property
    def inputs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """{name:index}"""
    @property
    def outs_info(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MPTypeProto]:
        """"""
    @property
    def attrs(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___AttrProto]: ...
    doc_string: builtins.str
    def __init__(
        self,
        *,
        op_type: builtins.str = ...,
        name: builtins.str = ...,
        inputs: collections.abc.Iterable[builtins.str] | None = ...,
        outs_info: collections.abc.Iterable[global___MPTypeProto] | None = ...,
        attrs: collections.abc.Mapping[builtins.str, global___AttrProto] | None = ...,
        doc_string: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attrs", b"attrs", "doc_string", b"doc_string", "inputs", b"inputs", "name", b"name", "op_type", b"op_type", "outs_info", b"outs_info"]) -> None: ...

global___NodeProto = NodeProto

@typing_extensions.final
class VersionInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAJOR_FIELD_NUMBER: builtins.int
    MINOR_FIELD_NUMBER: builtins.int
    PATCH_FIELD_NUMBER: builtins.int
    EXTRA_FIELD_NUMBER: builtins.int
    major: builtins.int
    """Major version, incompatible changes"""
    minor: builtins.int
    """Minor version, backward compatible changes"""
    patch: builtins.int
    """Patch version, bug fixes"""
    extra: builtins.str
    """Extra information, like "beta", "rc1", etc."""
    def __init__(
        self,
        *,
        major: builtins.int = ...,
        minor: builtins.int = ...,
        patch: builtins.int = ...,
        extra: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["extra", b"extra", "major", b"major", "minor", b"minor", "patch", b"patch"]) -> None: ...

global___VersionInfo = VersionInfo

@typing_extensions.final
class GraphProto(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class AttrsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___AttrProto: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___AttrProto | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    VERSION_FIELD_NUMBER: builtins.int
    NODES_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    ATTRS_FIELD_NUMBER: builtins.int
    @property
    def version(self) -> global___VersionInfo: ...
    @property
    def nodes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___NodeProto]: ...
    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """{name:index}"""
    @property
    def attrs(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___AttrProto]: ...
    def __init__(
        self,
        *,
        version: global___VersionInfo | None = ...,
        nodes: collections.abc.Iterable[global___NodeProto] | None = ...,
        outputs: collections.abc.Iterable[builtins.str] | None = ...,
        attrs: collections.abc.Mapping[builtins.str, global___AttrProto] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["version", b"version"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["attrs", b"attrs", "nodes", b"nodes", "outputs", b"outputs", "version", b"version"]) -> None: ...

global___GraphProto = GraphProto



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from mplang.analysis import diagram as analysis
from mplang.core.cluster import ClusterSpec
from mplang.core.mpir import Writer
from mplang.core.tracer import TraceContext, TracedFunction, trace
from mplang.simp import prandint, reveal, sealFrom, srun


def _toy():  # simple function for viz
    x = prandint(0, 10)
    y = prandint(0, 10)
    x_ = sealFrom(x, 0)
    y_ = sealFrom(y, 1)
    z_ = srun(lambda a, b: a < b)(x_, y_)
    z = reveal(z_)
    return z


def test_sequence_diagram_basic():
    cluster = ClusterSpec.simple(2)
    tctx = TraceContext(cluster)
    traced: TracedFunction = trace(tctx, _toy)
    graph = Writer().dumps(traced.make_expr())
    diagram = analysis.to_sequence_diagram(graph, world_size=2)
    assert "sequenceDiagram" in diagram
    # Accept any backend-specific naming for sealing and revealing primitives.
    seal_tokens = ("seal", "makeshare", "makeshares")
    reveal_tokens = ("reveal", "reconstruct")
    assert any(t in diagram for t in seal_tokens), diagram
    assert any(t in diagram for t in reveal_tokens), diagram


def _dummy_fn(a, b):
    # a simple function to trace; multiplication ensures an op appears
    return a + b * 2


def test_dump_markdown_sections(tmp_path):
    # Trace the function using analysis internal mechanisms via to_flowchart (side effect: requires a traced fn)
    # We indirectly obtain a traced function by using the diagram utilities which accept python callables.
    # However, dump expects a traced function object; importing here to avoid broad public surface changes.
    # Build a simple MPContext and trace manually to avoid importing heavy frontends
    cluster = ClusterSpec.simple(1)
    tctx = TraceContext(cluster)
    traced: TracedFunction = trace(tctx, _dummy_fn, 1, 2)

    md_path = tmp_path / "report.md"
    mpir_path = tmp_path / "graph.mpir"
    result = analysis.dump(
        traced,
        cluster_spec=cluster,
        sequence=True,
        flow=True,
        include_ir=True,
        report_path=str(md_path),
        mpir_path=str(mpir_path),
        title="Test Report",
        seq_opts={"collapse_local": True},
        flow_opts={"direction": "LR"},
    )

    content = result.markdown

    # Load from file to ensure write path executed
    written = md_path.read_text(encoding="utf-8")
    assert written == content

    # Core sections
    assert "## Compiler IR (text)" in content
    assert "## Graph Structure Analysis" in content

    # Mermaid fenced blocks (at least one sequence and one flowchart)
    mermaid_blocks = re.findall(r"```mermaid[\s\S]*?```", content)
    assert len(mermaid_blocks) >= 1

    # Ensure mpir file was written and is non-empty
    assert mpir_path.exists()
    assert mpir_path.read_text(encoding="utf-8").strip() != ""

    # Structured result fields sanity
    assert result.ir is not None
    assert result.sequence is not None
    assert result.flow is not None
    assert "sequenceDiagram" in result.sequence



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd

from mplang.core.pfunc import PFunction
from mplang.core.table import TableType
from mplang.core.tensor import TensorType
from mplang.kernels.base import cur_kctx, kernel_def
from mplang.runtime.data_providers import get_provider, resolve_uri
from mplang.utils import table_utils


def _to_numpy(obj: Any) -> np.ndarray:  # minimal helper to avoid duplicating logic
    if isinstance(obj, np.ndarray):
        return obj
    if hasattr(obj, "numpy"):
        try:
            return np.asarray(obj.numpy())  # type: ignore
        except Exception:
            pass
    return np.asarray(obj)


@kernel_def("builtin.identity")
def _identity(pfunc: PFunction, value: Any) -> Any:
    # Runtime guarantees exactly one argument; no extra arity checks here.
    return value


@kernel_def("builtin.read")
def _read(pfunc: PFunction) -> Any:
    path = pfunc.attrs.get("path")
    if path is None:
        raise ValueError("missing path attr for builtin.read")
    out_t = pfunc.outs_info[0]
    uri = resolve_uri(str(path))
    prov = get_provider(uri.scheme)
    if prov is None:
        raise NotImplementedError(f"no resource provider for scheme: {uri.scheme}")
    ctx = cur_kctx()
    try:
        return prov.read(uri, out_t, ctx=ctx)
    except Exception as e:  # pragma: no cover - provider errors
        raise RuntimeError(f"builtin.read failed: {e}") from e


@kernel_def("builtin.write")
def _write(pfunc: PFunction, obj: Any) -> Any:
    path = pfunc.attrs.get("path")
    if path is None:
        raise ValueError("missing path attr for builtin.write")
    uri = resolve_uri(str(path))
    prov = get_provider(uri.scheme)
    if prov is None:
        raise NotImplementedError(f"no resource provider for scheme: {uri.scheme}")
    ctx = cur_kctx()
    try:
        prov.write(uri, obj, ctx=ctx)
        return obj
    except Exception as e:  # pragma: no cover
        raise RuntimeError(f"builtin.write failed: {e}") from e


@kernel_def("builtin.constant")
def _constant(pfunc: PFunction) -> Any:
    data_bytes = pfunc.attrs.get("data_bytes")
    if data_bytes is None:
        raise ValueError("missing data_bytes attr for builtin.constant")
    out_t = pfunc.outs_info[0]
    fmt = pfunc.attrs.get("data_format")
    if isinstance(out_t, TableType):
        if fmt != "bytes[csv]":
            raise ValueError(f"unsupported table constant format {fmt}")
        df = table_utils.csv_to_dataframe(data_bytes)
        return df
    # tensor path
    shape = out_t.shape  # type: ignore[attr-defined,union-attr]
    dtype = out_t.dtype.numpy_dtype()  # type: ignore[attr-defined,union-attr]
    arr = np.frombuffer(data_bytes, dtype=dtype).reshape(shape)
    return arr


@kernel_def("builtin.rank")
def _rank(pfunc: PFunction) -> Any:
    ctx = cur_kctx()
    return np.array(ctx.rank, dtype=np.uint64)


@kernel_def("builtin.prand")
def _prand(pfunc: PFunction) -> Any:
    shape = pfunc.attrs.get("shape", ())
    rng = np.random.default_rng()
    info = np.iinfo(np.uint64)
    data = rng.integers(
        low=info.min, high=info.max, size=shape, dtype=np.uint64, endpoint=True
    )
    return data


@kernel_def("builtin.table_to_tensor")
def _table_to_tensor(pfunc: PFunction, table: Any) -> Any:
    if not isinstance(table, pd.DataFrame):
        raise TypeError("expected pandas DataFrame")
    if table.shape[1] == 0:
        raise ValueError("cannot pack empty table")
    mat = np.column_stack([table[col].to_numpy() for col in table.columns])
    return mat


@kernel_def("builtin.tensor_to_table")
def _tensor_to_table(pfunc: PFunction, tensor: Any) -> Any:
    arr = _to_numpy(tensor)
    if arr.ndim != 2:
        raise ValueError("tensor_to_table expects rank-2 array")
    col_names = pfunc.attrs.get("column_names")
    if col_names is None:
        raise ValueError("missing column_names attr")
    df = pd.DataFrame(arr, columns=list(col_names))
    return df


def _summ(v: Any) -> str:
    try:
        if isinstance(v, pd.DataFrame):
            return str(v.head(8).to_string(index=False))
        arr = _to_numpy(v)
        return str(
            np.array2string(
                arr, threshold=64, edgeitems=3, precision=6, suppress_small=True
            )
        )
    except Exception as e:  # pragma: no cover
        return f"<unprintable {type(v).__name__}: {e}>"


@kernel_def("builtin.debug_print")
def _debug_print(pfunc: PFunction, val: Any) -> Any:
    prefix = pfunc.attrs.get("prefix", "")
    ctx = cur_kctx()
    print(f"[debug_print][rank={ctx.rank}] {prefix}{_summ(val)}")
    return val


@kernel_def("builtin.pack")
def _pack(pfunc: PFunction, value: Any) -> Any:
    outs_info = pfunc.outs_info
    if len(outs_info) != 1:
        raise ValueError("builtin.pack expects single output type")
    out_ty = outs_info[0]
    if not isinstance(out_ty, TensorType):
        raise TypeError("builtin.pack must return TensorType")
    if out_ty.dtype.numpy_dtype() != np.uint8:
        raise TypeError("builtin.pack output dtype must be uint8")

    if isinstance(value, pd.DataFrame):
        csv_bytes = table_utils.dataframe_to_csv(value)
        return np.frombuffer(csv_bytes, dtype=np.uint8)

    arr = _to_numpy(value)
    return np.frombuffer(arr.tobytes(order="C"), dtype=np.uint8)


@kernel_def("builtin.unpack")
def _unpack(pfunc: PFunction, packed: Any) -> Any:
    outs_info = pfunc.outs_info
    if len(outs_info) != 1:
        raise ValueError("builtin.unpack expects single output type")
    out_ty = outs_info[0]

    b = np.asarray(packed, dtype=np.uint8).reshape(-1)

    if isinstance(out_ty, TensorType):
        np_dtype = out_ty.dtype.numpy_dtype()
        shape = tuple(out_ty.shape)
        if any(dim < 0 for dim in shape):
            raise ValueError("builtin.unpack does not support dynamic tensor shapes")
        elem_count = int(np.prod(shape))
        expected = elem_count * np.dtype(np_dtype).itemsize
        if b.size != expected:
            raise ValueError(
                f"unpack size mismatch: got {b.size} bytes, expect {expected} for {np_dtype} {shape}"
            )
        arr = np.frombuffer(b.tobytes(), dtype=np_dtype)
        return arr.reshape(shape)

    if isinstance(out_ty, TableType):
        csv_bytes = b.tobytes()
        return table_utils.csv_to_dataframe(csv_bytes)

    raise TypeError("builtin.unpack output type must be TensorType or TableType")



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for PHE frontend."""

import pytest

from mplang.core.dtype import FLOAT32, INT32
from mplang.ops import phe
from tests.ops.dummy import DummyTensor


def test_mul_validation():
    """Test PHE multiplication validation logic."""

    # Test that float x float is blocked (requires truncation)
    # Under strict typed_op semantics positional args must be MPObject or TypeSpecs.
    # We validate purely at type level here using scalar TensorType placeholders.
    float_ct = DummyTensor(FLOAT32, ())
    float_pt = DummyTensor(FLOAT32, ())

    with pytest.raises(
        ValueError,
        match="PHE multiplication does not support floating-point plaintext",
    ):
        phe.mul(float_ct, float_pt)

    # Test that float x int is allowed (no truncation required)
    int_pt = DummyTensor(INT32, ())

    # This should not raise a validation error (may fail for other reasons like missing keys)
    try:
        phe.mul(float_ct, int_pt)
    except ValueError as e:
        # Should not be the float x float validation error
        assert "float x float operations" not in str(e)
    except Exception:
        # Other exceptions are acceptable for this validation test
        pass

    # Test that int x float is allowed (no truncation required)
    int_ct = DummyTensor(INT32, ())

    try:
        phe.mul(int_ct, float_pt)
    except ValueError as e:
        # Should not be the float x float validation error
        assert "float x float operations" not in str(e)
    except Exception:
        # Other exceptions are acceptable for this validation test
        pass



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import numpy as np

import mplang
import mplang.simp as simp
from mplang.core import TensorType
from mplang.ops import builtin


@mplang.function
def save_data():
    # Party 0 creates and saves data
    x = simp.constant(np.array([[1, 2], [3, 4]], dtype=np.float32))
    y = simp.constant(np.array([[5, 6], [7, 8]], dtype=np.float32))

    x = simp.runAt(0, builtin.write)(x, path="tmp/x.npy")
    y = simp.runAt(1, builtin.write)(y, path="tmp/y.npy")

    return x, y


@mplang.function
def load_data():
    tensor_info = TensorType(shape=(2, 2), dtype=jnp.float32)

    x = simp.runAt(0, builtin.read)(path="tmp/x.npy", ty=tensor_info)
    y = simp.runAt(1, builtin.read)(path="tmp/y.npy", ty=tensor_info)

    x_ = simp.sealFrom(x, 0)
    y_ = simp.sealFrom(y, 1)
    z_ = simp.srun(lambda a, b: a + b)(x_, y_)
    z = simp.reveal(z_)

    return z


def run_stdio_example():
    print("\n--- Session 1: Creating and saving data ---")
    print("Creating matrices:")
    print("x = [[1, 2], [3, 4]]")
    print("y = [[5, 6], [7, 8]]")

    sim2 = mplang.Simulator.simple(2)
    _ = mplang.evaluate(sim2, save_data)
    print("Data saved to files successfully")

    print("\n--- Session 2: Loading data and computing x + y ---")
    sim3 = mplang.Simulator.simple(3)
    z = mplang.evaluate(sim3, load_data)
    result = mplang.fetch(sim3, z)
    print(result)


if __name__ == "__main__":
    run_stdio_example()



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging

import mplang.core.primitive as prim
from mplang.core import Mask, MPObject, Rank, function


# scatter :: [m a] -> m Rank -> m a
@function
def scatter_m(to_mask: Mask, root: Rank, args: list[MPObject]) -> MPObject:
    """Scatter the object from root to the parties in pmask.

    Args:
        to_mask: The mask of the parties that will receive the object.
        root: The rank of the root party.
        args: The objects to be scattered, which must hold by root and length of pmask'ed parties.
    """
    # sanity check, ensure all args are in the to_mask.
    for arg in args:
        if arg.pmask is None:
            logging.warning(f"Scattering dynamic {arg} from static root {root}")
        else:
            if not Mask.from_ranks(root).is_subset(arg.pmask):
                raise ValueError(f"Expect root {root} in {arg.pmask}, got {arg}.")

    to_ranks = list(Mask(to_mask))
    if len(args) != len(to_ranks):
        raise ValueError(f"Expect {len(to_ranks)} args, got {len(args)}. ")

    scattered = [
        prim.pshfl_s(arg, Mask.from_ranks(to_rank), [root])
        for to_rank, arg in zip(to_ranks, args, strict=False)
    ]

    result = prim.pconv(scattered)
    assert result.pmask == to_mask, (result.pmask, to_mask)
    return result  # type: ignore[no-any-return]


# gather :: m a -> m Rank -> [m a]
@function
def gather_m(src_mask: Mask, root: Rank, arg: MPObject) -> list[MPObject]:
    """Gather the object from pmask'ed parties to the root party.

    Args:
        src_pmask: The mask of the parties that will gather the object.
        root: The rank of the root party.
        arg: The object to be gathered, which must be the subset of pmask.

    Returns:
        A list of objects, with length equal to the number of parties in pmask.
    """
    # static pmask check.
    if arg.pmask is None:
        logging.warning(f"Gathering {arg} from {src_mask}, may raise RuntimeError.")
    else:
        if not Mask(src_mask).is_subset(arg.pmask):
            raise ValueError(f"Expect {src_mask} in {arg.pmask}, got {arg}.")

    result = []
    root_mask = Mask.from_ranks(root)
    for src_rank in Mask(src_mask):
        # Shuffle data from src_rank to root
        gathered_data = prim.pshfl_s(arg, root_mask, [src_rank])
        result.append(gathered_data)

    assert len(result) == Mask(src_mask).num_parties(), (result, src_mask)
    return result


# bcast :: m a -> m Rank -> m a
@function
def bcast_m(pmask: Mask, root: Rank, obj: MPObject) -> MPObject:
    """Broadcast the object from the root party to the parties in pmask."""
    if obj.pmask is None:
        logging.warning(f"Broadcasting {obj} from {root}, may raise RuntimeError.")
    else:
        if not Mask.from_ranks(root).is_subset(obj.pmask):
            raise ValueError(f"Expect root {root} in obj mask {obj.pmask}.")

    result = prim.pshfl_s(obj, pmask, [root] * Mask(pmask).num_parties())

    assert result.pmask == pmask, (result.pmask, pmask)
    return result  # type: ignore[no-any-return]


# p2p :: m Rank -> m Rank -> m a -> m a
@function
def p2p(frm: Rank, to: Rank, obj: MPObject) -> MPObject:
    """Point-to-point communication from frm to to."""

    # sanity check, ensure the object is in the frm mask.
    if obj.pmask is None:
        logging.warning(f"P2P {obj} from {frm} to {to}, may raise RuntimeError.")
    else:
        if not Mask.from_ranks(frm).is_subset(obj.pmask):
            raise ValueError(f"Expect {frm} in {obj.pmask}, got {obj}.")

    if frm == to:
        return obj

    return prim.pshfl_s(obj, Mask.from_ranks(to), [frm])  # type: ignore[no-any-return]


# allgather :: m a -> [m a]
@function
def allgather_m(pmask: Mask, arg: MPObject) -> list[MPObject]:
    """Gather the object from all parties in pmask and return a list of objects."""

    if arg.pmask is None:
        logging.warning(f"Allgathering {arg} from {pmask}, may raise RuntimeError.")
    else:
        if not Mask(pmask).is_subset(arg.pmask):
            raise ValueError(f"Expect {pmask} in {arg.pmask}, got {arg}.")

    # TODO(jint): implement me.
    raise NotImplementedError("Allgather not implemented")



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from mplang.core.dtype import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    FLOAT16,
    FLOAT32,
    FLOAT64,
    INT8,
    INT16,
    INT32,
    INT64,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    DType,
)
from mplang.core.mask import Mask
from mplang.core.mpir import dtype_to_proto
from mplang.core.mptype import MPType
from mplang.core.tensor import TensorType

try:
    import jax.numpy as jnp

    JAX_AVAILABLE = True
except ImportError:
    JAX_AVAILABLE = False


class TestDType:
    """Test cases for the DType class."""

    def test_dtype_creation(self):
        """Test basic DType creation and validation."""
        # Test valid boolean dtype
        bool_dtype = DType("bool", 1, True, False, False)
        assert bool_dtype.name == "bool"
        assert bool_dtype.bitwidth == 1
        assert bool_dtype.is_signed
        assert not bool_dtype.is_floating
        assert not bool_dtype.is_complex

        # Test valid integer dtype
        int32_dtype = DType("int32", 32, True, False, False)
        assert int32_dtype.name == "int32"
        assert int32_dtype.bitwidth == 32
        assert int32_dtype.is_signed
        assert not int32_dtype.is_floating
        assert not int32_dtype.is_complex

        # Test valid float dtype
        float64_dtype = DType("float64", 64, True, True, False)
        assert float64_dtype.name == "float64"
        assert float64_dtype.bitwidth == 64
        assert float64_dtype.is_signed
        assert float64_dtype.is_floating
        assert not float64_dtype.is_complex

        # Test valid complex dtype
        complex128_dtype = DType("complex128", 128, True, True, True)
        assert complex128_dtype.name == "complex128"
        assert complex128_dtype.bitwidth == 128
        assert complex128_dtype.is_signed
        assert complex128_dtype.is_floating
        assert complex128_dtype.is_complex

    def test_dtype_constants(self):
        """Test predefined dtype constants."""
        # Test boolean
        assert BOOL.name == "bool"
        assert BOOL.bitwidth == 8  # NumPy bool is 8 bits

        # Test integers
        assert INT8.name == "int8"
        assert INT8.bitwidth == 8
        assert INT64.name == "int64"
        assert INT64.bitwidth == 64

        # Test unsigned integers
        assert UINT32.name == "uint32"
        assert UINT32.bitwidth == 32
        assert not UINT32.is_signed

        # Test floats
        assert FLOAT32.name == "float32"
        assert FLOAT32.bitwidth == 32
        assert FLOAT32.is_floating

        # Test complex
        assert COMPLEX64.name == "complex64"
        assert COMPLEX64.bitwidth == 64
        assert COMPLEX64.is_complex

    def test_from_numpy(self):
        """Test creating DType from NumPy dtypes."""
        # Test various numpy dtypes
        test_cases = [
            (np.bool_, BOOL),
            (np.int8, INT8),
            (np.int16, INT16),
            (np.int32, INT32),
            (np.int64, INT64),
            (np.uint8, UINT8),
            (np.uint16, UINT16),
            (np.uint32, UINT32),
            (np.uint64, UINT64),
            (np.float16, FLOAT16),
            (np.float32, FLOAT32),
            (np.float64, FLOAT64),
            (np.complex64, COMPLEX64),
            (np.complex128, COMPLEX128),
        ]

        for np_dtype, expected_dtype in test_cases:
            result = DType.from_numpy(np_dtype)
            assert result == expected_dtype

        # Test with numpy dtype objects
        result = DType.from_numpy(np.dtype("float32"))
        assert result == FLOAT32

    def test_from_python_type(self):
        """Test creating DType from Python types."""
        assert DType.from_python_type(bool) == BOOL
        assert DType.from_python_type(int) == INT64
        assert DType.from_python_type(float) == FLOAT64
        assert DType.from_python_type(complex) == COMPLEX128

        # Test with unsupported type
        with pytest.raises(ValueError):
            DType.from_python_type(str)

    def test_from_any(self):
        """Test creating DType from various inputs."""
        # From numpy types
        assert DType.from_any(np.float32) == FLOAT32
        assert DType.from_any(np.dtype("int64")) == INT64

        # From python types
        assert DType.from_any(float) == FLOAT64
        assert DType.from_any(int) == INT64

        # From DType (should return same)
        assert DType.from_any(FLOAT32) == FLOAT32

        # From string
        assert DType.from_any("float64") == FLOAT64

        # Test with unsupported type
        with pytest.raises(ValueError):
            DType.from_any("unsupported")

    def test_to_numpy(self):
        """Test converting DType to NumPy dtype."""
        test_cases = [
            (BOOL, np.bool_),
            (INT32, np.int32),
            (UINT64, np.uint64),
            (FLOAT32, np.float32),
            (COMPLEX128, np.complex128),
        ]

        for dtype, expected_np_type in test_cases:
            result = dtype.to_numpy()
            assert result == np.dtype(expected_np_type)

    def test_to_python_type(self):
        """Test converting DType to Python type."""
        assert BOOL.to_python_type() is bool
        assert INT64.to_python_type() is int
        assert FLOAT64.to_python_type() is float
        assert COMPLEX128.to_python_type() is complex

        # Unsigned integers should map to int
        assert UINT32.to_python_type() is int

    @pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX not available")
    def test_to_jax(self):
        """Test converting DType to JAX dtype."""
        test_cases = [
            (BOOL, jnp.bool_),
            (INT32, jnp.int32),
            (FLOAT32, jnp.float32),
            (COMPLEX64, jnp.complex64),
        ]

        for dtype, expected_jax_type in test_cases:
            result = dtype.to_jax()
            assert result == expected_jax_type

    def test_equality(self):
        """Test DType equality comparison."""
        # Same dtypes should be equal
        assert FLOAT32 == FLOAT32
        assert INT64 == INT64

        # Different dtypes should not be equal
        assert FLOAT32 != FLOAT64
        assert INT32 != UINT32

        # Equal dtypes created separately should be equal
        dtype1 = DType("float32", 32, True, True, False)
        dtype2 = DType("float32", 32, True, True, False)
        assert dtype1 == dtype2

    def test_repr(self):
        """Test DType string representation."""
        assert repr(FLOAT32) == "DType('float32')"
        assert repr(INT64) == "DType('int64')"
        assert repr(BOOL) == "DType('bool')"

    def test_hash(self):
        """Test DType hashing."""
        # Same dtypes should have same hash
        assert hash(FLOAT32) == hash(FLOAT32)

        # Different dtypes should have different hashes (usually)
        assert hash(FLOAT32) != hash(FLOAT64)

        # Can be used in sets and dicts
        dtype_set = {FLOAT32, INT64, BOOL}
        assert len(dtype_set) == 3

    def test_dtype_conversion_chain(self):
        """Test conversion chain between different representations."""
        # Start with DType
        dtype = FLOAT64

        # Convert to numpy and back
        np_dtype = dtype.to_numpy()
        back_to_dtype = DType.from_numpy(np_dtype)
        assert dtype == back_to_dtype

        # Convert to python type and back
        py_type = dtype.to_python_type()
        from_py_dtype = DType.from_python_type(py_type)
        assert dtype == from_py_dtype

    def test_mpir_dtype_to_proto_compatibility(self):
        """Test that mpir.dtype_to_proto works with new DType objects."""
        # Create TensorType with DType
        tensor_info = TensorType(FLOAT32, (2, 2))

        # dtype_to_proto should work with DType objects
        proto_result = dtype_to_proto(tensor_info.dtype)

        # Should be the same as using numpy dtype directly
        proto_result_np = dtype_to_proto(np.float32)
        assert proto_result == proto_result_np


class TestTensorInfo:
    """Test cases for the TensorType class."""

    def test_tensor_info_creation(self):
        """Test TensorType compatibility with custom dtypes."""
        # Test with NumPy array
        arr = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        tensor_info = TensorType.from_obj(arr)

        # dtype should now return DType, not numpy dtype
        assert isinstance(tensor_info.dtype, DType)
        assert tensor_info.dtype == FLOAT32
        assert tensor_info.shape == (2, 3)

        # Use to_numpy() for compatibility
        assert tensor_info.to_numpy() == np.dtype("float32")

        # Test with scalar
        scalar_info = TensorType.from_obj(42)
        assert isinstance(scalar_info.dtype, DType)
        assert scalar_info.dtype == INT64
        assert scalar_info.shape == ()
        assert scalar_info.to_numpy() == np.dtype("int64")

        # Test with float scalar
        float_scalar_info = TensorType.from_obj(3.14)
        assert isinstance(float_scalar_info.dtype, DType)
        assert float_scalar_info.dtype == FLOAT64
        assert float_scalar_info.shape == ()
        assert float_scalar_info.to_numpy() == np.dtype("float64")

    def test_tensor_info_constructor_with_dtype(self):
        """Test TensorType constructor with various dtype inputs."""
        # Test constructing TensorType with custom DType
        tensor_info = TensorType(FLOAT32, (2, 3))
        assert isinstance(tensor_info.dtype, DType)
        assert tensor_info.dtype == FLOAT32
        assert tensor_info.to_numpy() == np.dtype("float32")

        # Test constructing with numpy dtype (should be converted)
        tensor_info_np = TensorType(np.dtype("int32"), (5,))
        assert isinstance(tensor_info_np.dtype, DType)
        assert tensor_info_np.dtype == INT32
        assert tensor_info_np.to_numpy() == np.dtype("int32")

    def test_tensor_info_unified_api(self):
        """Test the unified dtype API for TensorType."""
        # Test TensorType
        tensor_info = TensorType(np.float32, (3, 4))

        # dtype should return DType
        assert isinstance(tensor_info.dtype, DType)
        assert tensor_info.dtype.name == "float32"

        # to_numpy() for explicit conversion
        assert tensor_info.to_numpy() == np.dtype("float32")

    def test_dtype_consistency(self):
        """Test that dtype and to_numpy() are consistent."""
        # Create from DType
        tensor_info = TensorType(FLOAT32, (2, 2))

        # The numpy dtype should match what dtype can produce
        assert tensor_info.to_numpy() == tensor_info.dtype.to_numpy()

        # Create from numpy dtype
        tensor_info2 = TensorType(np.int16, (3, 3))
        assert tensor_info2.dtype == INT16

    def test_dtype_enhanced_functionality(self):
        """Test that dtype provides enhanced functionality."""
        tensor_info = TensorType(np.float32, (2, 2))
        dtype = tensor_info.dtype

        # Enhanced properties
        assert dtype.to_numpy() == np.float32
        assert dtype.to_python_type() is float
        if JAX_AVAILABLE:
            assert dtype.to_jax() == jnp.float32

        # Type properties
        assert dtype.is_floating
        assert not dtype.is_complex
        assert dtype.bitwidth == 32

    def test_edge_cases(self):
        """Test edge cases and error conditions."""
        # Test with unusual numpy dtypes
        tensor_info = TensorType(np.dtype("<f4"), (2, 2))  # Little-endian float32
        assert tensor_info.dtype == FLOAT32

        # Test hash consistency
        dtype_set = {tensor_info.dtype, FLOAT32}
        assert len(dtype_set) == 1  # Should be the same


class TestMPType:
    """Test cases for the MPType class."""

    def test_mp_type_creation(self):
        """Test MPType compatibility with custom dtypes."""
        # Test with tensor
        arr = np.array([1, 2, 3, 4], dtype=np.int64)
        mp_type = MPType.from_tensor(arr, pmask=Mask.from_int(0b111))

        # dtype should now return DType, not numpy dtype
        assert isinstance(mp_type.dtype, DType)
        assert mp_type.dtype == INT64
        assert mp_type.shape == (4,)
        assert mp_type.pmask == Mask.from_int(0b111)

        # Use to_numpy() for compatibility
        assert mp_type.to_numpy() == np.dtype("int64")

        # Test with scalar
        scalar_mp_type = MPType.from_tensor(42.0, pmask=Mask.from_int(0b101))
        assert isinstance(scalar_mp_type.dtype, DType)
        assert scalar_mp_type.dtype == FLOAT64
        assert scalar_mp_type.shape == ()
        assert scalar_mp_type.pmask == Mask.from_int(0b101)
        assert scalar_mp_type.to_numpy() == np.dtype("float64")

        # Test with custom attributes
        attr_mp_type = MPType.from_tensor(
            arr, pmask=Mask.from_int(0b111), encrypted=True
        )
        assert attr_mp_type.attrs == {"encrypted": True}

    def test_mp_type_constructor_with_dtype(self):
        """Test MPType constructor with various dtype inputs."""
        # Test constructing MPType with custom DType
        mp_type = MPType.tensor(FLOAT32, (2, 3), 0b111)
        assert isinstance(mp_type.dtype, DType)
        assert mp_type.dtype == FLOAT32
        assert mp_type.to_numpy() == np.dtype("float32")

        # Test constructing with numpy dtype (should be converted)
        mp_type_np = MPType.tensor(INT32, (5,), 0b101)
        assert isinstance(mp_type_np.dtype, DType)
        assert mp_type_np.dtype == INT32
        assert mp_type_np.to_numpy() == np.dtype("int32")

    def test_mp_type_unified_api(self):
        """Test the unified dtype API for MPType."""
        # Test MPType
        mp_type = MPType.tensor(INT64, (2, 2), 0b111)

        # dtype should return DType
        assert isinstance(mp_type.dtype, DType)
        assert mp_type.dtype.name == "int64"

        # to_numpy() for explicit conversion
        assert mp_type.to_numpy() == np.dtype("int64")

    def test_array_creation_workflow(self):
        """Test typical array creation and manipulation workflow."""
        # Create tensor info from numpy array
        arr = np.random.random((3, 4)).astype(np.float32)
        tensor_info = TensorType.from_obj(arr)

        # Verify dtype information
        assert tensor_info.dtype == FLOAT32
        assert tensor_info.dtype.is_floating
        assert tensor_info.dtype.bitwidth == 32

        # Create compatible numpy array
        new_arr = np.zeros((2, 2), dtype=tensor_info.to_numpy())
        assert new_arr.dtype == np.float32

        # Create MPType with same dtype
        mp_type = MPType.tensor(tensor_info.dtype, (5, 5), 0b111)
        assert mp_type.dtype == tensor_info.dtype



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, runtime_checkable

import numpy as np

from mplang.core.dtype import DType

# basic type aliases
Shape = tuple[int, ...]
ScalarType = int | float | bool | complex

__all__ = ["ScalarType", "Shape", "TensorLike", "TensorType"]


@runtime_checkable
class TensorLike(Protocol):
    """
    Protocol for objects structurally resembling tensors from common libraries
    (NumPy, PyTorch, JAX), focusing on dtype and shape attributes.
    """

    @property
    def dtype(self) -> Any: ...

    @property
    def shape(self) -> Shape: ...


@dataclass(frozen=True)
class TensorType:
    """A data class that describes the type information of a tensor."""

    dtype: DType
    shape: Shape

    def __init__(self, dtype: DType | Any, shape: Shape):
        # Convert dtype to DType if needed
        if not isinstance(dtype, DType):
            dtype = DType.from_any(dtype)
        object.__setattr__(self, "dtype", dtype)
        object.__setattr__(self, "shape", shape)

    @classmethod
    def from_obj(cls, obj: TensorLike | ScalarType) -> TensorType:
        if isinstance(obj, ScalarType):
            return cls(DType.from_python_type(type(obj)), ())
        elif isinstance(obj, TensorLike):
            return cls(DType.from_any(obj.dtype), obj.shape)
        else:
            raise TypeError(f"Unsupported type: {type(obj)}.")

    def to_numpy(self) -> np.dtype:
        """Convert to NumPy dtype for compatibility."""
        return self.dtype.to_numpy()

    def __repr__(self) -> str:
        shape_str = "x".join(str(d) for d in self.shape)
        dtype_name = str(self.dtype)
        return f"Tensor<{shape_str}x{dtype_name}>" if shape_str else f"{dtype_name}"



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import Any

import cloudpickle as pickle
import spu.libspu as libspu

from mplang.core.comm import ICollective, ICommunicator


class LinkCommunicator(ICommunicator, ICollective):
    """Wraps libspu link communicator for distributed communication"""

    def __init__(self, rank: int, addrs: list[str], *, mem_link: bool = False):
        self._rank = rank
        self._world_size = len(addrs)

        desc = libspu.link.Desc()  # type: ignore
        desc.recv_timeout_ms = 100 * 1000  # 100 seconds
        desc.http_max_payload_size = 32 * 1024 * 1024  # Default set link payload to 32M
        for rank, addr in enumerate(addrs):
            desc.add_party(f"P{rank}", addr)

        if mem_link:
            self.lctx = libspu.link.create_mem(desc, self._rank)
        else:
            self.lctx = libspu.link.create_brpc(desc, self._rank)

        logging.info(
            f"LinkCommunicator initialized with rank={self._rank}, world_size={self._world_size}, addrs={addrs}",
        )

        self._counter = 0

    @property
    def rank(self) -> int:
        return self.lctx.rank  # type: ignore[no-any-return]

    @property
    def world_size(self) -> int:
        return self.lctx.world_size  # type: ignore[no-any-return]

    def get_lctx(self) -> libspu.link.Context:
        """Get the link context"""
        return self.lctx

    # override
    def new_id(self) -> str:
        res = self._counter
        self._counter += 1
        return str(res)

    def wrap(self, obj: Any) -> str:
        data = pickle.dumps(obj)
        return data.hex()  # type: ignore[no-any-return]

    def unwrap(self, obj: str) -> Any:
        data = bytes.fromhex(obj)
        return pickle.loads(data)  # type: ignore[no-any-return]

    def send(self, to: int, key: str, data: Any) -> None:
        serialized = pickle.dumps((key, data))
        self.lctx.send(to, serialized.hex())

    def recv(self, frm: int, key: str) -> Any:
        serialized = self.lctx.recv(frm)
        rkey, data = pickle.loads(bytes.fromhex(serialized.decode()))
        assert key == rkey, f"recv key {key} != {rkey}"
        return data  # type: ignore[no-any-return]

    def p2p(self, frm: int, to: int, data: Any) -> Any:
        assert 0 <= frm < self.world_size
        assert 0 <= to < self.world_size

        # TODO: link handles cid internally?
        cid = self.new_id()

        if self.rank == frm:
            self.send(to, cid, data)
            return None
        elif self.rank == to:
            return self.recv(frm, cid)
        else:
            return None

    def gather(self, root: int, data: Any) -> list[Any]:
        assert 0 <= root < self.world_size
        rets = self.lctx.gather(self.wrap(data), root)
        return [self.unwrap(ret) for ret in rets]

    def scatter(self, root: int, args: list[Any]) -> Any:
        assert 0 <= root < self.world_size
        assert len(args) == self.world_size, f"{len(args)} != {self.world_size}"
        ret = self.lctx.scatter([self.wrap(arg) for arg in args], root)
        return self.unwrap(ret)

    def allgather(self, arg: Any) -> list[Any]:
        rets = self.lctx.all_gather(self.wrap(arg))
        return [self.unwrap(ret) for ret in rets]

    def bcast(self, root: int, arg: Any) -> Any:
        assert 0 <= root < self.world_size
        ret = self.lctx.broadcast(self.wrap(arg), root)
        return self.unwrap(ret)

    def gather_m(self, pmask: int, root: int, data: Any) -> list[Any]:
        raise ValueError("Not supported by LinkCommunicator")

    def scatter_m(self, pmask: int, root: int, args: list[Any]) -> Any:
        raise ValueError("Not supported by LinkCommunicator")

    def allgather_m(self, pmask: int, arg: Any) -> list[Any]:
        raise ValueError("Not supported by LinkCommunicator")

    def bcast_m(self, pmask: int, root: int, arg: Any) -> Any:
        raise ValueError("Not supported by LinkCommunicator")



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Integration tests for mock crypto pipeline (frontend -> runtime -> backend)
# Marked as integration due to use of Simulator and multi-party coordination.

from __future__ import annotations

import numpy as np
import pytest

import mplang
import mplang.simp as simp
from mplang.core.tensor import TensorType
from mplang.ops import builtin, crypto

pytestmark = pytest.mark.integration


def _mk_sim(n: int = 3):
    return mplang.Simulator.simple(n)


def test_keygen_shape():
    sim = _mk_sim()

    @mplang.function
    def fn():
        k16 = simp.runAt(0, crypto.keygen)(16)
        k32 = simp.runAt(1, crypto.keygen)()
        return k16, k32

    k16, k32 = mplang.evaluate(sim, fn)
    a16, a32 = mplang.fetch(sim, (k16, k32))
    assert a16[0].shape == (16,)
    assert a32[1].shape == (32,)
    assert a16[0].dtype == np.uint8


def test_enc_dec_roundtrip_bytes():
    sim = _mk_sim()

    @mplang.function
    def fn():
        key = simp.runAt(0, crypto.keygen)(32)
        pt = simp.runAt(0, lambda: np.arange(50, dtype=np.uint8))()
        ct = simp.runAt(0, crypto.enc)(pt, key)
        rt = simp.runAt(0, crypto.dec)(ct, key)
        return pt, rt

    pt, rt = mplang.evaluate(sim, fn)
    pt_v, rt_v = mplang.fetch(sim, (pt, rt))
    np.testing.assert_array_equal(pt_v[0], rt_v[0])


def test_pack_unpack_roundtrip_various():
    sim = _mk_sim()

    shapes_dtypes = [
        ((), np.int32),
        ((4,), np.float32),
        ((2, 3), np.int64),
    ]

    @mplang.function
    def fn():
        outs = []
        for _idx, (shape, dt) in enumerate(shapes_dtypes):
            arr = simp.runAt(0, lambda s=shape, d=dt: np.zeros(s, dtype=d))()
            packed = simp.runAt(0, builtin.pack)(arr)
            unpacked = simp.runAt(0, builtin.unpack)(
                packed, out_ty=TensorType.from_obj(arr)
            )
            outs.append(unpacked)
        return tuple(outs)

    (u0, u1, u2) = mplang.evaluate(sim, fn)
    (v0, v1, v2) = mplang.fetch(sim, (u0, u1, u2))
    assert v0[0].shape == ()
    assert v1[0].shape == (4,)
    assert v2[0].shape == (2, 3)


def test_kem_hkdf_symmetric_mock():
    sim = _mk_sim()

    @mplang.function
    def fn():
        sk0, pk0 = simp.runAt(0, crypto.kem_keygen)()
        sk1, pk1 = simp.runAt(1, crypto.kem_keygen)()
        pk1_on_0 = simp.p2p(1, 0, pk1)
        pk0_on_1 = simp.p2p(0, 1, pk0)
        sec0 = simp.runAt(0, crypto.kem_derive)(sk0, pk1_on_0)
        sec1 = simp.runAt(1, crypto.kem_derive)(sk1, pk0_on_1)
        k0 = simp.runAt(0, crypto.hkdf)(sec0, "info")
        k1 = simp.runAt(1, crypto.hkdf)(sec1, "info")
        return k0, k1

    k0, k1 = mplang.evaluate(sim, fn)
    v0, v1 = mplang.fetch(sim, (k0, k1))
    np.testing.assert_array_equal(v0[0], v1[1])


@pytest.mark.parametrize("length", [1, 7, 32, 64])
def test_variable_key_lengths(length: int):
    sim = _mk_sim()

    @mplang.function
    def fn(n: int):
        return simp.runAt(0, crypto.keygen)(n)

    k = mplang.evaluate(sim, fn, length)
    v = mplang.fetch(sim, k)
    assert v[0].shape == (length,)
    assert v[0].dtype == np.uint8



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import jax
import jax.numpy as jnp
import pytest
from jax.tree_util import tree_flatten, tree_unflatten

from mplang.core.expr.evaluator import create_evaluator
from mplang.core.pfunc import PFunction
from mplang.core.tensor import TensorType
from mplang.kernels import stablehlo  # noqa: F401
from mplang.kernels.context import RuntimeContext
from mplang.ops import jax_cc

# Enable 64-bit precision in JAX for testing different dtypes
jax.config.update("jax_enable_x64", True)


class TestStablehloKernel:
    """Tests for mlir.stablehlo flat backend kernel."""

    @pytest.fixture(autouse=True)
    def _evaluator(self):
        # Minimal single-rank communicator (no actual messaging needed for local kernel)
        from mplang.core.comm import CommunicatorBase

        class _SingleComm(CommunicatorBase):  # type: ignore[misc]
            def send(self, to: int, key: str, data):  # pragma: no cover - unused
                raise RuntimeError("send should not be called in single-rank test")

        comm = _SingleComm(rank=0, world_size=1)
        runtime = RuntimeContext(rank=0, world_size=1)
        ev = create_evaluator(rank=0, env={}, comm=comm, runtime=runtime)
        self.ev = ev
        yield

    @pytest.mark.parametrize(
        "test_function, inputs",
        [
            # simple_function
            (
                lambda x, y: x + y * 2,
                (jnp.array([1.0, 2.0, 3.0]), jnp.array([0.5, 1.5, 2.5])),
            ),
            # multi_output_function
            (
                lambda x, y: (x + y, x * y),
                (jnp.array([2.0, 3.0]), jnp.array([1.0, 4.0])),
            ),
            # complex_function_with_math_ops
            (
                lambda x, y, z: (
                    (
                        c := jnp.matmul(
                            (x + y).reshape(-1, 1), jnp.sin(z).reshape(1, -1)
                        )
                    ),
                    jnp.sum(c),
                ),
                (
                    jnp.array([1.0, 2.0]),
                    jnp.array([0.5, 1.5]),
                    jnp.array([0.1, 0.2]),
                ),
            ),
            # compilation_preserves_function_signature
            (
                lambda a, b, c: a * b + c,
                (
                    jnp.array([[1.0, 2.0], [3.0, 4.0]]),
                    jnp.array([[0.5, 1.5], [2.5, 3.5]]),
                    jnp.array([[0.1, 0.2], [0.3, 0.4]]),
                ),
            ),
            # empty_input_function
            (lambda: jnp.array([1.0, 2.0, 3.0]), ()),
            # scalar_function
            (
                lambda x, y: jnp.sum(x * y),
                (jnp.array([1.0, 2.0, 3.0]), jnp.array([4.0, 5.0, 6.0])),
            ),
            # matrix_operations
            (
                lambda a, b: jnp.transpose(jnp.matmul(a, b)) + 1.0,
                (
                    jnp.array([[1.0, 2.0], [3.0, 4.0]]),
                    jnp.array([[5.0, 6.0], [7.0, 8.0]]),
                ),
            ),
            # different_dtypes
            (
                lambda x, y: x + y,
                (
                    jnp.array([1.0, 2.0], dtype=jnp.float32),
                    jnp.array([3.0, 4.0], dtype=jnp.float64),
                ),
            ),
        ],
    )
    def test_function_execution(self, test_function, inputs):
        """Verify end-to-end compilation and execution pipeline for diverse function types."""
        # Establish ground truth via local JAX execution
        expected = test_function(*inputs)

        # Generate tensor metadata for runtime input validation
        [TensorType(shape=x.shape, dtype=x.dtype) for x in inputs]

        # Compile function to portable StableHLO MLIR representation
        is_var = lambda obj: hasattr(obj, "dtype") and hasattr(obj, "shape")
        cfunc, _, out_tree = jax_cc.jax2stablehlo(is_var, test_function, *inputs)

        # Execute via evaluator (dispatches kernel by fn_type)
        result_flat = self.ev._exec_pfunc(cfunc, list(inputs))  # type: ignore[attr-defined]

        # Reconstruct nested output structure from flattened results
        result = tree_unflatten(out_tree, result_flat)

        # Validate numerical correctness within tolerance bounds
        expected_flat, _ = tree_flatten(expected)
        result_flat_from_tree, _ = tree_flatten(result)
        assert len(result_flat_from_tree) == len(expected_flat)
        for res, exp in zip(result_flat_from_tree, expected_flat, strict=False):
            assert jnp.allclose(jnp.asarray(res), jnp.asarray(exp), rtol=1e-5)

    def test_invalid_format_execution(self):
        invalid_pfunc = PFunction(
            fn_type="invalid_format",
            ins_info=(),
            outs_info=(),
            fn_name="test",
            fn_text="invalid_text",
        )
        from mplang.core.comm import CommunicatorBase

        class _SingleComm(CommunicatorBase):  # type: ignore[misc]
            def send(self, to: int, key: str, data):  # pragma: no cover - unused
                raise RuntimeError("send should not be called in single-rank test")

        comm = _SingleComm(rank=0, world_size=1)
        runtime = RuntimeContext(rank=0, world_size=1)
        ev = create_evaluator(0, {}, comm, runtime)
        with pytest.raises(NotImplementedError):
            ev._exec_pfunc(invalid_pfunc, [])  # type: ignore[attr-defined]



# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import warnings

import numpy as np
from numpy.typing import NDArray

from mplang.core.pfunc import PFunction
from mplang.kernels.base import cur_kctx, kernel_def

__all__: list[str] = []


def _rng() -> np.random.Generator:
    kctx = cur_kctx()
    rt = kctx.runtime
    r = rt.get_state("tee.rng")
    if r is None:
        seed = int(os.environ.get("MPLANG_TEE_SEED", "0")) + kctx.rank * 10007
        r = np.random.default_rng(seed)
        rt.set_state("tee.rng", r)
    assert isinstance(r, np.random.Generator)  # type narrowing for mypy
    return r


def _quote_from_pk(pk: np.ndarray) -> NDArray[np.uint8]:
    header = np.array([1], dtype=np.uint8)
    pk32 = np.asarray(pk, dtype=np.uint8).reshape(32)
    out: NDArray[np.uint8] = np.concatenate([header, pk32]).astype(np.uint8)  # type: ignore[assignment]
    return out


@kernel_def("mock_tee.quote_gen")
def _tee_quote_gen(pfunc: PFunction, pk: object) -> NDArray[np.uint8]:
    warnings.warn(
        "Insecure mock TEE kernel 'mock_tee.quote_gen' in use. NOT secure; for local testing only.",
        stacklevel=3,
    )
    pk = np.asarray(pk, dtype=np.uint8)
    # rng access ensures deterministic seeding per rank even if unused now
    _rng()
    return _quote_from_pk(pk)


@kernel_def("mock_tee.attest")
def _tee_attest(pfunc: PFunction, quote: object) -> NDArray[np.uint8]:
    warnings.warn(
        "Insecure mock TEE kernel 'mock_tee.attest' in use. NOT secure; for local testing only.",
        stacklevel=3,
    )
    quote = np.asarray(quote, dtype=np.uint8)
    platform = pfunc.attrs.get("platform")
    if platform is None:
        raise ValueError("missing required 'platform' attribute in PFunction")

    if quote.size != 33:
        raise ValueError("mock quote must be 33 bytes (1 header + 32 pk)")
    return quote[1:33].astype(np.uint8)
